grammar VhdlRTL;
 
options {
  k         = 2;
  backtrack = true;
  output    = template;
  rewrite   = true;
}
 
@header {
      package com.eu.miscedautils.vhdlparser;
      import com.eu.miscedautils.CommonUtils.LogMessages;
 
      import java.util.Iterator;
      import java.util.List;
      import java.util.ArrayList;
      import java.util.LinkedList;
      import org.antlr.runtime.ANTLRFileStream;
    import org.antlr.runtime.ANTLRStringStream;
    import org.antlr.runtime.CharStream;
    import org.antlr.runtime.CommonTokenStream;
    import org.antlr.runtime.RecognitionException;
    import org.antlr.runtime.TokenSource;
    import org.antlr.runtime.TokenStream;
    import java.io.File;
    import java.io.FileInputStream;
    import java.io.IOException;
}
 
@members
{
    String workLib = null;
    LinkedList<String> uses;
    LinkedList<String> libs;
    boolean isInSubProgramScope = false;
    FVhPInstantiation currInstance = null;
    FVhPBase currSubprog = null;
    FVhPComponentSpecification currCompSpec  = null;
    FVhPBase myCurrentTarget = null;
    FVhPBase myCurrentSind = null;
	public void displayRecognitionError(String[] tokenNames,
                                        RecognitionException e) {
        String hdr = getErrorHeader(e);
        String msg = getErrorMessage(e, tokenNames);
        LogMessages.instance().print(3554, LogMessages.MSG_SEVERITY.ERROR, 
                                     FVhPDescription.instance().getFile() + ", " + hdr + " : " + msg );
    }

    
    LinkedList<FVhPGeneric> getCurrentInstanceMasterGenericList()
    {
    	LinkedList<FVhPGeneric> generics = null;
    	if ( null == currInstance && null == currCompSpec )
    	{
    		LogMessages.instance().print(3553, LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,"Cuurent instance is NULL while retrieving master's generic list  ...");
    		return null;                                
    	}
    	FVhPBase master = null;
    	master = ( null != currInstance ) ? currInstance.getMaster() : currCompSpec.getComponent();
    	
    	switch( master.getObjType() )
    	{
    		case COMPONENTDECL:
    			generics = ((FVhPComponentDecl)master).getGenerics();
    			break;
    		case ENTITY:
    			generics = ((FVhPEntity)master).getGenerics();
    			break;	
    		default:
    			LogMessages.instance().print(3552, LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,"Unsupported object type " +
    			                              "found as master for retrieving generic list ...");
    			break;                               
    	}
    	return generics;
    }
    
    
    LinkedList<FVhPPort> getCurrentInstanceMasterPortList()
    {
    	LinkedList<FVhPPort> ports = null;
    	if ( null == currInstance && null == currCompSpec )
    	{
    		LogMessages.instance().print(3551, LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,
    		                             "Current instance is NULL while retrieving master port list ...");
    		return null;  
    	}
    	FVhPBase master = ( null != currInstance ) ? currInstance.getMaster() : currCompSpec.getComponent();
    	switch( master.getObjType() )
    	{
    		case COMPONENTDECL:
    			ports = ((FVhPComponentDecl)master).getPorts();
    			break;
    		case ENTITY:
    			ports = ((FVhPEntity)master).getPorts();
    			break;
    		default:
    			LogMessages.instance().print(3550, 
    			                             LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,
    			                             "Unsupported object type " +
    			                              "found as master for retrieving port list ...");
    			break;                               
    	}
    	return ports;
    }
        
    FVhPGeneric getCurrentInstanceFormalGeneric(String name)
    {
    	FVhPGeneric generic = null;
    	if ( null == currInstance && null == currCompSpec )
    	{
    		LogMessages.instance().print(3549,
    		                             LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,
    		                             "Current instance is NULL while retrieving formal generic  ...");  
 			return null;  
    	}
    	FVhPBase master = ( null != currInstance ) ? currInstance.getMaster() : currCompSpec.getComponent();   	
    	switch( master.getObjType() )
    	{
    		case COMPONENTDECL:
    			generic = ((FVhPComponentDecl)master).getGeneric(name);
    			break;
    		case ENTITY:
    			generic = ((FVhPEntity)master).getGeneric(name);
    			break;	
    		default:
    			LogMessages.instance().print(3548,
    			                             LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,
    			                             "Unsupported object type " +
    			                              "found as master for retrieving generic ...");
    			break;                               
    	}
    	return generic;
    }
    
    
    FVhPBase getCurrentInstanceFormalPort(String name)
    {
    	FVhPBase port = null;
    	if ( null == currInstance && null == currCompSpec && null == currSubprog )
    	{
    		LogMessages.instance().print(3547, LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,"Current instance is NULL while retrieving formal port ...");  
 			return null;  
    	}
    	FVhPBase master = ( null != currInstance ) ? currInstance.getMaster() : 
    	                  ( ( null != currSubprog ) ? currSubprog : currCompSpec.getComponent() );
    	LinkedList<FVhPInterfaceElement> itfs = null;
    	switch( master.getObjType() )
    	{
    		case COMPONENTDECL:
    			port = ((FVhPComponentDecl)master).getPort(name);
    			break;
    		case ENTITY:
    			port = ((FVhPEntity)master).getPort(name);
    			break;
    		case SUBPROGRAM_BODY:
    		    itfs = ((FVhPSubProgramBody)master).getInterfaceElementList();
    			break;
    		case SUBPROGRAM_DECL:
    			itfs = ((FVhPSubProgramDecl)master).getInterfaceElementList();
    			break;    				
    		default:
    			LogMessages.instance().print(3546, LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,"Unsupported object type " +
    			                              "found as master for retrieving port ...");
    			break;                               
    	}
    	if ( null != itfs && false == itfs.isEmpty() )
    	{
    		Iterator<FVhPInterfaceElement> itr = itfs.iterator();
    		while ( itr.hasNext() )
    		{
    			FVhPInterfaceElement intf = itr.next();
    			if ( intf.getName().equalsIgnoreCase( name ) )
    			{
    				port = intf;
    				break;
    			}
    		}
    	}
    	return port;
    }
    
    String getNameString(FVhPBase object)
    {
        if ( null == object )
        {
            return null;
        }
        String str = null;
        switch(object.getObjType())
        {
            case SIMPLENAME:
                str = ((FVhPSimpleName)object).getName();
                break;   
             case SELECTEDNAME:
                str = ((FVhPSelectedName)object).print();
                break;      
            default:
                LogMessages.instance().print(3545,
                 LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,
                 "Unsupported object type '" + 
                 object.getObjType().toString() + 
                 "' found for getNameString() for construct at line " + object.getLine() + " ...");
                break;    
        }
        return str;
    }
    
        Long getLongValue(String value)
        {
        	Long longval = null;
        	if (null == value )
        	{
        		return longval;
        	}
        	if ( value.length() < 20 )
        	{
        		longval = Long.parseLong( value );
        	}
        	else if ( value.length() == 20 )
        	{
        		String tmp = value.substring(0, 18);
        		longval = Long.parseLong( tmp );
        		String tmp2 = value.substring(19);
        		longval = longval * 10 +  Long.parseLong( tmp2 );
        	}
        	else
        	{
        		LogMessages.instance().print(3544,
        		 LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR,"Failed to process long values since digit count exceeds 20 : '" + value + "' ..."); 
        	}
        	return longval;
        }
        
        public void setCurrentTarget( FVhPBase target )
        {
            myCurrentTarget = target;
        }
        
        public FVhPBase getCurrentTarget()
        {
            return myCurrentTarget;
        }
        
        public FVhPRecordElement getRecordFiledFromCurrentTargetType(String field)
        {
            if ( null == myCurrentTarget && null == myCurrentSind )
            {
                return null;
            }
            FVhPSubTypeIndication sind = null;
            if ( null != myCurrentSind )
            {
                sind = (FVhPSubTypeIndication)myCurrentSind;
            }
            else
            {
                sind = getSignalOrVariableOrConstSind( myCurrentTarget );
            }
            if ( null == sind )
            {
                return null;
            }
            FVhPBase typemark = sind.getTypeMark();
            if ( null == typemark )
            {
                return null;
            }
            if ( typemark instanceof FVhPSubTypeDecl )
            {
                while ( null != typemark && typemark instanceof FVhPSubTypeDecl )
                {
                    typemark = ((FVhPSubTypeDecl)typemark).getSubtypeIndication();
                    typemark = ((FVhPSubTypeIndication)typemark).getTypeMark();
                }
            }
             if ( typemark instanceof FVhPTypeDecl )
            {
                typemark  = ((FVhPTypeDecl)typemark).getActualDef();
            }
            if ( false == ( typemark instanceof FVhPRecordType ) )
            {
                return null;
            }
            FVhPRecordElement ele = ((FVhPRecordType)typemark).getElement( field );
            return ele;
        }
        
        public FVhPSubTypeIndication getSignalOrVariableOrConstSind( FVhPBase obj )
        {
            FVhPSubTypeIndication sind = null;
            if ( null == obj )
            {
                return null;
            }
            switch ( obj.getObjType() )
            {
                case VARIABLE :
                {
                    sind = ((FVhPVariable)obj).getType();
                    break;
                }
                case SIGNAL :
                {
                    sind = ((FVhPSignal)obj).getType();
                    break;
                }
                case CONSTANT :
                {
                    sind = ((FVhPConstant)obj).getType();
                    break;
                }
            }
            return sind;
        }
}

@lexer::header {
     package com.eu.miscedautils.vhdlparser;
      
     import java.util.Iterator;
      import java.util.List;
      import java.util.ArrayList;
      import java.util.LinkedList;
 
}

@lexer::members
{
    String workLib = null;
    LinkedList<String> libs;
    boolean isInSubProgramScope = false;
}
 
 
parseVhdlRTL [String wrklib, boolean isSTDPack] returns [ FVhPRoot root]
    @init
    {
        workLib = wrklib;
        FVhPDescription.instance().setCurrentWorkLib(workLib);
        this.uses = new LinkedList<String>();
        this.libs = new LinkedList<String>();
        if ( false == isSTDPack  )
        {
        	// We need to always load the STD library and STANDARD package except 
        	// when the library name is STD
            try 
            {
                FVhPBase du = FVhPDescription.instance().restoreDesignUnit("std","standard");
                FVhPBase lib = FVhPDescription.instance().getSymbol( "work" );
                if ( null == lib || FVhPBase.ObjType.LIBRARY != lib.getObjType() )
                {
                    lib = new FVhPLibrary(-1, "work");
                    FVhPDescription.instance().addSymbol( "work", lib );
                }
                lib = FVhPDescription.instance().getSymbol( workLib);
                if ( null == lib || FVhPBase.ObjType.LIBRARY != lib.getObjType() )
                {
                    lib = new FVhPLibrary(-1, workLib);
                    FVhPDescription.instance().addSymbol( workLib, lib );
                }
//                FVhPDescription.instance().getRoot().addDesignUnit( du );
            }
            catch (IOException e) 
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
        }
    }
    :
  (design_file
    {
        $root = (FVhPRoot)$design_file.desc.getRoot();
    }
  )*  EOF
  ;


abstract_literal returns [FVhPExprBase expr]
    @init
    {
        String literal = null;
    }
  :
  declit=DECIMAL_LITERAL
  {
    Integer line = $declit.line;
    literal = ( null == $declit ) ? null : $declit.getText().replaceAll("_","");
    String []temp1 = literal.split("\\.");
    Long intpart = getLongValue(temp1[0]);
//    Long intpart = longpart.intValue();    
//    Integer intpart = Integer.parseInt(temp1[0]);
    Long decimalpart = null;
    Integer exponentpart = null;
    Character sign = null;
    if ( temp1.length > 1 )  
    {
        String []temp2 = temp1[1].split("e");
        if ( temp2.length == 1 )
        {
            temp2 = temp1[1].split("E");
        }
        decimalpart = getLongValue(temp2[0]);
        $expr = new FVhPExprDecimalLiteral( line, intpart );
        ((FVhPExprDecimalLiteral)$expr).setDecimalPart( decimalpart );        
        if ( temp2.length > 1 )
        {
            sign = temp2[1].charAt(0);
            if ( '+' == sign || '-' == sign )
            {
                exponentpart = Integer.parseInt(temp2[1].substring(1));
            }
            else
            {
                sign = '+';
                exponentpart = Integer.parseInt(temp2[1].substring(0));
            }
            ((FVhPExprDecimalLiteral)$expr).setExponentPart( exponentpart );
            ((FVhPExprDecimalLiteral)$expr).setExponentSign( sign );    
        }
    }
    else
    {
        String []temp2 = literal.split("e");
        if ( temp2.length == 1 )
        {
            temp2 = literal.split("E");
        }
        if ( temp2.length > 1 && null != temp2[1] )
        {
            $expr = new FVhPExprDecimalLiteral( line, intpart );
            sign = temp2[1].charAt(0);
            if ( '+' == sign || '-' == sign )
            {
                exponentpart = Integer.parseInt(temp2[1].substring(1));
            }
            else
            {
                sign = '+';
                exponentpart = Integer.parseInt(temp2[1].substring(0));
            }
            ((FVhPExprDecimalLiteral)$expr).setExponentPart( exponentpart );
            ((FVhPExprDecimalLiteral)$expr).setExponentSign( sign );
        }
        else
        {
            $expr = new FVhPExprInteger( line, intpart.intValue());
        }
    }
  }
  | basedlit=BASED_LITERAL
  {
    Integer line = $BASED_LITERAL.line;
    literal = ( null == $basedlit ) ? null : $basedlit.getText().replaceAll("_","");
    String []hashSplit = literal.split("#");
    Integer base = Integer.parseInt(hashSplit[0]);
    String []temp1 = hashSplit[1].split("\\.");    
    String intpart = temp1[0];
    String decimalpart = null;
    Integer exponentpart = null;
    Character sign = null;
    if ( temp1.length > 1 )  
    {
        decimalpart = temp1[1];
    }
    $expr = new FVhPExprBasedDecimalLiteral( line, base, intpart );
    ((FVhPExprBasedDecimalLiteral)$expr).setDecimalPart( decimalpart );
    if ( hashSplit.length > 2 && null != hashSplit[2])
    {
        if ( 'e' == hashSplit[2].charAt(0) || 'E' == hashSplit[2].charAt(0) )
        {
            sign = hashSplit[2].charAt(1);
            if ( '+' == sign || '-' == sign )
            {
                exponentpart = Integer.parseInt(hashSplit[2].substring(2));
            }
            else
            {
                sign = '+';
                exponentpart = Integer.parseInt(hashSplit[2].substring(1));
            }
            ((FVhPExprBasedDecimalLiteral)$expr).setExponentSign( sign );
            ((FVhPExprBasedDecimalLiteral)$expr).setExponentPart( exponentpart );
        }
    }
  }
  ;


access_type_definition returns[FVhPTypeDefBase accesstype]
  :
  ACCESS subtype_indication
  {
  	$accesstype = new FVhPAccessType( $ACCESS.getLine(), $subtype_indication.sind );
  }
  ;

actual_designator returns[FVhPExprBase actual]
  :
  expression
  {
     $actual = $expression.expr;
  }  
  | OPEN
  {
  	 $actual = new FVhPExprOpen($OPEN.getLine());
  }  
  ;

actual_parameter_part returns[LinkedList<FVhPExprBase> paramlist]
  :
  association_list
  {
  	$paramlist = $association_list.actuals;
  }  
  ;

actual_part returns[FVhPExprBase actual]
  :
  (name LPAREN) => nm1=name
  {
       LogMessages.instance().print(3601, LogMessages.MSG_SEVERITY.WARNING, "unsupported actual type ...");
  }
  LPAREN actual_designator RPAREN
  {
     LogMessages.instance().print(3600, LogMessages.MSG_SEVERITY.WARNING, "Actual part is not supported yet ...");
  } 
  | 
  actual_designator
  {
     $actual = $actual_designator.actual;
  }  
  ;


adding_operator returns [FVhPExprBinary.OpType op]
  :
  PLUS { $op = FVhPExprBinary.OpType.PLUS; }
  | MINUS  { $op = FVhPExprBinary.OpType.MINUS; }
  | AMPERSAND { $op = FVhPExprBinary.OpType.AMPERSAND; }
  ;

aggregate returns [FVhPAggregate aggr]
  :
  LPAREN assoc1=element_association 
  {
      $aggr = new FVhPAggregate( $assoc1.eleassoc.getLine(), null );
      $aggr.addElement ( $assoc1.eleassoc );
  }
  (COMMA 
      assoc2=element_association
      {
          $aggr.addElement ( $assoc2.eleassoc );
      }
  )* RPAREN
  ;

alias_declaration returns [FVhPBase decl]
  :
  ALIAS alias_designator (COLON alias_indication)? IS name (signature)? SEMI
  {
    FVhPAliasDeclaration aliasdecl = new  FVhPAliasDeclaration( $alias_designator.tok.getLine(),
                                                                $alias_designator.tok.getText(),
                                                                $alias_indication.sind);
                                                                
    FVhPBase actualObj = FVhPDescription.instance().getSymbol( $name.tok.getText() );                                                      
    aliasdecl.setActualObj( actualObj );
    aliasdecl.setRangelist ( $name.rangelist );
    if ( null != $name.attribName || $name.isBitString )
    {
        LogMessages.instance().print(3599, LogMessages.MSG_SEVERITY.ERROR, "Missing some ddetails on alias processing ..." ); 
    }
    FVhPDescription.instance().addDeclaration( aliasdecl );
    FVhPDescription.instance().addSymbol( $alias_designator.tok.getText(), aliasdecl );
    $decl = aliasdecl;
  }
  ;

alias_designator returns [Token tok, boolean isIdentifier, boolean isCharLit, boolean isStringLit]
  :
  identifier
  {
      $tok = $identifier.tok;
      $isIdentifier = true;
  }  
  | 
  CHARACTER_LITERAL
  {
    $tok = $CHARACTER_LITERAL;
      $isCharLit = true;
  }  
  | 
  STRING_LITERAL
  {
    $tok = $STRING_LITERAL;
      $isStringLit = true;
  }  
  ;

alias_indication returns [FVhPSubTypeIndication sind]
  :
  subtype_indication
  {
      $sind = $subtype_indication.sind;
  }
  ;

allocator returns [FVhPExprBase expr]
  :
  NEW
  (
    (name APOSTROPHE) => qualified_expression
    {
        $expr = new FVhPExprAllocator( $NEW.getLine(), $qualified_expression.expr );
    }
    | s1=subtype_indication
    {
        $expr = new FVhPExprAllocator( $NEW.getLine(), $s1.sind );
    }
  )
//  {
//     LogMessages.instance().print(3598, LogMessages.MSG_SEVERITY.WARNING, "allocator is not supported yet ...");
//  }  
  ;

//architecture_body
//  : ARCHITECTURE identifier OF name IS
//    architecture_declarative_part
//    BEGIN
//    architecture_statement_part
//    END ( ARCHITECTURE )? ( identifier )? SEMI
//  ;
 
architecture_body returns [FVhPArchitecture arch]
  :
  ARCHITECTURE id1=identifier OF nm1=name IS
  {
      String archname = $id1.tok.getText();
      String entName = getNameString($nm1.obj);
      $arch = new FVhPArchitecture(archname, entName, $id1.tok.getLine());
      FVhPDescription desc = FVhPDescription.instance();
      desc.insertSymbol( archname, $arch );
      FVhPEntity ent = null;
      try
      {
      	ent = (FVhPEntity)desc.restoreDesignUnit( workLib, entName );
      }
      catch (IOException e) 
      {
		// TODO Auto-generated catch block
    	e.printStackTrace();
      }
      $arch.setEntity( ent );
      desc.pushScope($arch);
  } 
  architecture_declarative_part 
  BEGIN architecture_statement_part END (ARCHITECTURE)? (identifier)? SEMI
  {
    FVhPDescription.instance().popScope();
    FVhPDescription.instance().getRoot().addDesignUnit($arch);
    libs = null;
    uses = null;
  }
  ;

architecture_declarative_part
  :
  (block_declarative_item)*
  ;

architecture_statement
  :
  block_statement
  {
     LogMessages.instance().print(3597, LogMessages.MSG_SEVERITY.WARNING, "block statement is not supported yet ...");
  }  
  |
  (
    (label_colon)? (POSTPONED)? PROCESS
  )
    => process_statement
  |
  (
    (label_colon)? (POSTPONED)? procedure_call SEMI
  )
    => concurrent_procedure_call_statement
  |
  (
    (label_colon)? (POSTPONED)? ASSERT
  )
    => concurrent_assertion_statement
  |
  (
    (label_colon)? (POSTPONED)?
    (
      conditional_signal_assignment
      | selected_signal_assignment
    )
  )
    => concurrent_signal_assignment_statement
  | (label_colon instantiated_unit) => component_instantiation_statement
  | (label_colon generation_scheme GENERATE) => generate_statement
  | simultaneous_statement
  ;

architecture_statement_part
  :
  (architecture_statement)*
  ;

array_type_definition returns [FVhPTypeDefBase def]
  :
  (unconstrained_array_definition) => u1=unconstrained_array_definition
  {
      $def = $u1.def;
  }
  |
  constrained_array_definition
  {
      $def = $constrained_array_definition.def;
  }
  |
  ;
  
assertion
  :
  ASSERT condition (REPORT expression)? (SEVERITY expression)?
  {
     LogMessages.instance().print(3596, LogMessages.MSG_SEVERITY.WARNING, "assertion statement is not supported yet ...");
  }  
  ;

assertion_statement
  :
  (label_colon)? assertion SEMI
  ;

association_element returns [FVhPBase formalPort, FVhPBase formalGeneric, FVhPExprBase actual]
  :
  ( 
  (formal_part ARROW) => formal_part ARROW
  {
  	String portName = $formal_part.formal.getText();
  	$formalPort = getCurrentInstanceFormalPort(portName);
  	if ( null == $formalPort )
  	{
  		$formalGeneric = getCurrentInstanceFormalGeneric(portName);
  	}
  }
  )? 
  actual_part
  {
  	$actual = $actual_part.actual;
  }
  ;

association_list returns [LinkedList<FVhPPortMap> portmaps, 
                          LinkedList<FVhPGenericMap> genericmaps,
                          LinkedList<FVhPExprBase> actuals ]
  :
  assoc1=association_element
  {
  	if ( ( null != $assoc1.formalPort ) || ( null != $assoc1.formalGeneric ) )
  	{
  		if ( null != $assoc1.formalPort )
  		{
  			$portmaps = new LinkedList<FVhPPortMap>();
  			FVhPPortMap map = new FVhPPortMap( $assoc1.formalPort.getLine(), $assoc1.formalPort, $assoc1.actual);
  			$portmaps.add( map );
  		}
  		else 
  		{
  			$genericmaps = new LinkedList<FVhPGenericMap>();
  			FVhPGenericMap map = new FVhPGenericMap( $assoc1.formalGeneric.getLine(), $assoc1.formalGeneric, $assoc1.actual);
  			$genericmaps.add( map );
  		}
  	}
  	else
  	{
  		$actuals = new LinkedList<FVhPExprBase> ();
  		$actuals.add($assoc1.actual);
  	}
  }
  (
  COMMA assoc2=association_element
  {
  	if ( ( null != $assoc2.formalPort ) || ( null != $assoc2.formalGeneric ) )
  	{
  		if ( null != $assoc2.formalPort )
  		{
  			FVhPPortMap map = new FVhPPortMap( $assoc2.formalPort.getLine(), $assoc2.formalPort, $assoc2.actual);
  			$portmaps.add( map );
  		}
  		else 
  		{
			FVhPGenericMap map = new FVhPGenericMap( $assoc2.formalGeneric.getLine(), $assoc2.formalGeneric, $assoc2.actual);
  			$genericmaps.add( map );  	
  		}
  	}
  	else
  	{
  		if ( null == $actuals )
  		{
  			$actuals = new LinkedList<FVhPExprBase> ();
  		}
  		$actuals.add($assoc2.actual);
  	}
  }
  )*
  ;

attribute_declaration returns [FVhPBase attribdecl]
  :
  ATTRIBUTE label_colon name SEMI
  {
      String id = $name.tok.getText();
      FVhPBase sym = FVhPDescription.instance().getSymbol(id);
      $attribdecl = new FVhPAttributeDecl( $label_colon.tok.getText(), sym, $name.tok.getLine() );
     FVhPDescription.instance().addDeclaration( $attribdecl );
     FVhPDescription.instance().addSymbol( $label_colon.tok.getText(), $attribdecl );
  }  
  ;

// Need to add several tokens here, for they are both, VHDLAMS reserved words
// and attribute names.
// (25.2.2004, e.f.)

attribute_designator returns[String attribName]
  :
  identifier
  {
      $attribName = $identifier.tok.getText();
  }  
  | RANGETOK
  {
      $attribName = $RANGETOK.getText();
  }  
  ;

attribute_specification
  :
  ATTRIBUTE attribute_designator OF entity_specification IS expression SEMI
  {
     LogMessages.instance().print(3595, LogMessages.MSG_SEVERITY.WARNING,"attribute specification is not supported yet ...");
  }  
  ;

base_unit_declaration returns [FVhPBase baseunit]
  :
  identifier SEMI
  {
      $baseunit = FVhPDescription.instance().getSymbol( $identifier.tok.getText() );
      if ( null == $baseunit )
      {
          $baseunit = new FVhPPhysicalTypeBaseUnit( $identifier.tok.getLine(), $identifier.tok.getText() );
          FVhPDescription.instance().addSymbol( $identifier.tok.getText(), $baseunit );
      }
  }  
  ;

binding_indication returns [FVhPBindingIndication binding]
  :
  (
  USE entity_aspect
  {
  	// Load this entity
	String lib = null;
  	String entName = null;
 	FVhPBase primdu = null; 
 	FVhPBase secdu = null;	
  	if ( null != $entity_aspect.prim )
  	{
  		LinkedList<FVhPBase> eleList = ((FVhPSelectedName)$entity_aspect.prim).getElementList();
  		Iterator<FVhPBase> itr = eleList.iterator();
  		while ( itr.hasNext() )
  		{
  			FVhPBase tmp = itr.next();
  			String tmpName = null;
  			if ( null != tmp )
  			{
  				tmpName = tmp.print();
  			}
  			if ( null != entName && null != tmpName )
  			{
  				lib = entName;
  			}
  			entName = tmpName;
  		}
  		if ( null == lib )
  		{
  			lib = workLib;
  		}
  	}
  	else
  	{
  		lib = workLib;
  		entName = $entity_aspect.tok.getText();
  	}
  	if ( false == $entity_aspect.isOpen )
  	{
  		try 
    	{
    		primdu = FVhPDescription.instance().restoreDesignUnit(lib,entName);
    		if ( null != $entity_aspect.sec )
    		{
    			String archname = $entity_aspect.sec.getText();
    			secdu = FVhPDescription.instance().restoreDesignUnit(lib,entName+":"+archname);
    		}
    	}
    	catch (IOException e) 
    	{
    		// TODO Auto-generated catch block
    		e.printStackTrace();
    	}
    	$binding = new FVhPBindingIndication($USE.getLine(), primdu );
    	$binding.setLibName( lib );
    	$binding.setArchitecture( secdu );
    	$binding.setMasterDuType( $entity_aspect.duType );
    }
  }
  )? 
  (
  generic_map_aspect
  {
  	$binding.setGenericmaps( $generic_map_aspect.genericmaps ); 
  }
  )? 
  (
  port_map_aspect
  {
  	$binding.setPortmaps( $port_map_aspect.portmaps ); 
  }
  )?
  ;

block_configuration returns[FVhPBlockConfiguration blkcfg]
	@init
	{
		LinkedList<FVhPBase> configurationItems = null;
		FVhPBase currScope = null;
	}
  :
  f1=FOR b1=block_specification 
  {
  	$blkcfg = new FVhPBlockConfiguration( $f1.getLine(), $b1.blkName.getText() );
  	$blkcfg.setIndex( $b1.index );
  	FVhPBase scp = FVhPDescription.instance().getScope();
  	FVhPEntity ent = null;
  	if ( FVhPBase.ObjType.CONFIGURATIONDECLARATION == scp.getObjType() )
  	{
  	     ent = ((FVhPConfigurationDeclaration) scp).getEntity();
  	}
  	else if ( FVhPBase.ObjType.ENTITY == scp.getObjType() )
    {
        ent = (FVhPEntity) scp;
    } 
  	if ( null != ent )
  	{
  	     String wrk = FVhPDescription.instance().getCurrentWorkLib();
  	     try
  	     {
	  	     FVhPBase arch = FVhPDescription.instance().restoreDesignUnit( wrk, ent.getName() + ":" + $b1.blkName.getText() );
	  	     if ( null != arch )
	  	     {
	  	         currScope = arch;
	  	         FVhPDescription.instance().pushScope(arch);
	  	     }
  	     }
  	     catch (IOException e) 
         {
            e.printStackTrace();
         }
  	}
  }
  (use_clause
  {
  	LogMessages.instance().print(3594, LogMessages.MSG_SEVERITY.ERROR, "Use clause inside block_configuration is not supported yet ...");
  }
  )* 
  (
  c=configuration_item
  {
  	if ( null == configurationItems )
  	{
  		configurationItems = new LinkedList<FVhPBase>();
  	}
  	configurationItems.add( $c.cfg );
  }
  )* 
  {
  	$blkcfg.setConfigurationItems( configurationItems );
  }
  END FOR SEMI
  {
    if ( null != currScope )
    {
        FVhPDescription.instance().popScope();
        currScope = null;
    }
  } 
  ;

block_declarative_item
  :
  (subprogram_declaration) => subprogram_declaration
  | subprogram_body
  | type_declaration
  | subtype_declaration
  | constant_declaration
  | signal_declaration
  | variable_declaration
  | file_declaration
  | alias_declaration
  | component_declaration
  | attribute_declaration
  | attribute_specification
  | configuration_specification
  | disconnection_specification
  | use_clause
  | group_template_declaration
  | group_declaration
  ;

block_declarative_part
  :
  (block_declarative_item)*
  ;

block_header
  :
  (generic_clause (generic_map_aspect SEMI)?)? (port_clause (port_map_aspect SEMI)?)?
  {
	LogMessages.instance().print(3593, LogMessages.MSG_SEVERITY.WARNING, 
	                             "block header is not supported yet ...");
  }
  ;

block_specification returns[Token blkName, FVhPExprBase index]
  :
  (identifier) => identifier 
  {
  	$blkName = $identifier.tok;
  }
  (LPAREN 
  index_specification
  {
  	$index = $index_specification.expr;
  } 
  RPAREN)?
  | name 
  {
  	$blkName = $name.tok;
  }
  ;

block_statement returns [FVhPBase stmt]
	@init
	{
		FVhPBlockStatement blk = null;
	}
  :
  label_colon b1=BLOCK
  {
  	String label = $label_colon.tok.getText();
  	blk = new FVhPBlockStatement( $b1.getLine(), label);
  	$stmt = blk;
  	FVhPDescription.instance().addStatement( blk );
  	FVhPDescription.instance().pushScope( blk );
  } 
  (LPAREN expression
  {
    if ( null != blk )
    {
  	     blk.setCondition( $expression.expr );
  	}
  }
  RPAREN)? (IS)? 
  block_header block_declarative_part 
  BEGIN block_statement_part 
  END BLOCK (identifier)? SEMI
  {
  	 FVhPDescription.instance().popScope();
  }
  ;

block_statement_part
  :
  (architecture_statement)*
  ;

case_statement returns [ FVhPSeqCaseStatement casestmt]
    @init
    {
        String label = null;
    }
  :
  (label_colon { label = $label_colon.tok.getText();})? CASE expression IS 
  {
      $casestmt = new FVhPSeqCaseStatement( $expression.expr.getLine(), $expression.expr);
      $casestmt.setLabel(label);
      FVhPDescription.instance().addStatement( $casestmt );
      FVhPDescription.instance().pushScope( $casestmt );
  }
  (
      alt=case_statement_alternative
      {
          $casestmt.addItem( $alt.item );
      }
  )+ 
  END CASE (identifier)? SEMI
  {
      FVhPDescription.instance().popScope();
  }
  ;

case_statement_alternative returns [ FVhPSeqCaseItem item ]
  :
  WHEN choices ARROW
  {
      $item = new FVhPSeqCaseItem($WHEN.getLine(),$choices.choices );
      FVhPDescription.instance().pushScope($item);
  } 
  sequence_of_statements
  {
      FVhPDescription.instance().popScope();
  }
  ;

choice returns [FVhPBase choice]
  :
  (identifier) => id=identifier
  {
        FVhPDescription desc = FVhPDescription.instance();
        FVhPBase sym = (FVhPBase)desc.getSymbol( $id.tok.getText() );
        if ( null == sym )
        {
            sym = getRecordFiledFromCurrentTargetType( $id.tok.getText() );
        }
        if ( null == sym )
        {
            LogMessages.instance().print(3592, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+ 
                                       $id.tok.getText() + 
                                       "' in " +
                                       desc.getFile() +
                                       ":" + $id.tok.getLine() );
        } 
        else
        {
            $choice = new FVhPExprIdentifier( $id.tok.getLine(), sym );
        }      
  }
  | literal
  {
  	$choice = $literal.expr;
  }
  | (discrete_range) => disc=discrete_range
  {
      $choice = $disc.rng;
  }
  | simple_expression
  {
      $choice = $simple_expression.expr;
  }    
  | OTHERS
  {
      $choice = new FVhPExprOthers( $OTHERS.getLine() );
  }
  ;

choices returns [ LinkedList<FVhPBase> choices ]
  :
  c1=choice
  {
      $choices = new  LinkedList<FVhPBase>();
      $choices.add( $c1.choice );
  } 
  (
      BAR c2=choice
      {
          $choices.add( $c2.choice );
      }
  )*
  ;

component_configuration returns [FVhPComponentConfiguration compconfig]
    @init
    {
        FVhPBase currScope = null;
    }
  :
  f1=FOR c1=component_specification 
  {
  	$compconfig = new FVhPComponentConfiguration( $f1.getLine(), $c1.compspec );
  }
  (
  b1=binding_indication
  {
  	$compconfig.setBindingIndication( $b1.binding );
  	if ( null != $b1.binding  && null != $b1.binding.getMaster() )
  	{
  	 currScope = $b1.binding.getMaster();
  	 FVhPDescription.instance().pushScope( currScope );
  	}
  } 
  SEMI)? 
  (
  blk=block_configuration
  {
  	$compconfig.setBlockConfiguration($blk.blkcfg);
  }
  )? 
  END FOR SEMI
  {
    if ( null != currScope )
    {
        FVhPDescription.instance().popScope( );
        currScope = null;
    }
  }
  {
  	currCompSpec = null;
  }
  ;

//component_declaration
//  : COMPONENT identifier ( IS )?
//    ( generic_clause )?
//    ( port_clause )?
//    END COMPONENT ( identifier )? SEMI
//  ;

component_declaration returns [FVhPComponentDecl comp]
  :
  COMPONENT nm=identifier (IS)?
  {
  	$comp = new FVhPComponentDecl($nm.tok.getLine(), $nm.tok.getText());
  	FVhPDescription.instance().addSymbol( $nm.tok.getText(), $comp );
  	FVhPDescription.instance().addDeclaration($comp );
  	FVhPDescription.instance().pushScope( $comp );
  } 
  (
  generic_clause
  {
	$comp.setGenerics($generic_clause.generics);   
  }
  )? 
  (
  port_clause
  {
  	$comp.setPorts( $port_clause.ports );
  }
  )?
   END COMPONENT (nm2=identifier )? SEMI
  {
  	FVhPDescription.instance().popScope( );
  }
  ;

component_instantiation_statement returns[FVhPBase inst]
  :
  label_colon instantiated_unit 
  {
  	String instName = $label_colon.tok.getText();
  	FVhPBase master = $instantiated_unit.master;

  	$inst = new FVhPInstantiation( $label_colon.tok.getLine(), instName, master );
  	
  	currInstance = (FVhPInstantiation)$inst;
  	currInstance.setLibName($instantiated_unit.libName);
  	currInstance.setArchitecture($instantiated_unit.arch);
  	FVhPDescription.instance().addStatement($inst);
  }
  (
  generic_map_aspect
  {
  	((FVhPInstantiation)$inst).setGenericmaps( $generic_map_aspect.genericmaps ); 
  }
  )? 
  (
  port_map_aspect
  {
  	((FVhPInstantiation)$inst).setPortmaps( $port_map_aspect.portmaps ); 
  }
  )? SEMI
  {
  	currInstance = null;
  }
  ;

//component_specification
//  : instantiation_list COLON name
//  ;

component_specification returns[FVhPComponentSpecification compspec]
  :
  inst=instantiation_list COLON nm=name
  {
  	String compName = $name.tok.getText();
  	FVhPBase comp = FVhPDescription.instance().getSymbol( compName );
  	if ( null != comp )
  	{
  		//FVhPComponentDecl
  		$compspec = new FVhPComponentSpecification( $COLON.getLine(), comp );
  		$compspec.setInstanceNames( $inst.instNameList );
  		$compspec.setAll( $inst.isAll );
  		$compspec.setOthers( $inst.isOthers );
  		currCompSpec = $compspec;
  	}
  }
  ;

composite_nature_definition
  :
  record_nature_definition
  ;

composite_type_definition returns [ FVhPTypeDefBase def ]
  :
  array_type_definition
  {
      $def = $array_type_definition.def;
  }
  | 
  record_type_definition
  {
      $def = $record_type_definition.def;
  }
  ;

concurrent_assertion_statement
  :
  (label_colon)? (POSTPONED)? assertion SEMI
  ;

concurrent_procedure_call_statement returns[FVhPBase stmt]
	@init
	{
		String label = null;
	}
  :
  (label_colon {label = $label_colon.tok.getText();})? (POSTPONED)? procedure_call SEMI
  {
  	$stmt = $procedure_call.proccall;
  	if ( null != $stmt )
  	{
  		if ( FVhPBase.ObjType.INSTANTIATION == $stmt.getObjType() )
  		{
  			((FVhPInstantiation)$stmt).setInstanceName( label);
  		}
  		else
  		{
  			((FVhPSubProgramCall)$stmt).setLabel( label );
  		}
  		FVhPDescription.instance().addStatement( $stmt );
  	}
  } 
  ;

concurrent_signal_assignment_statement
  :
  (label_colon)? (POSTPONED)?
  (
    conditional_signal_assignment
    | selected_signal_assignment
  )
  ;

condition returns [FVhPExprBase expr]
  :
  expression
  {
      $expr = $expression.expr;
  }
  ;

condition_clause returns [FVhPExprBase expr]
  :
  UNTIL condition
  {
      $expr = $condition.expr;
  }
  ;
conditional_signal_assignment
  :
  target LE opts conditional_waveforms SEMI
  {
        FVhPBase assmt = null;
        if ( true == $conditional_waveforms.isConditional )
        {
            assmt = new FVhPConditionalAssignment(
                                                    $target.target.getLine(), 
                                                    $target.target, 
                                                    $conditional_waveforms.waveformList);
        }
        else
        {
            assmt = new FVhPSignalAssignment(
                                                    $target.target.getLine(), 
                                                    $target.target, 
                                                    $conditional_waveforms.waveformList);
        }                                                                             
        FVhPDescription.instance().addStatement(assmt);
  }
  ;

conditional_waveforms returns [LinkedList<FVhPWaveForm> waveformList, boolean isConditional]
    @init
    {
        LinkedList<FVhPWaveForm> wvfrms = null;
        FVhPWaveForm firstWvFrm = null;
        FVhPExprBase lastCond = null;
    }
    @after
    {
        $waveformList = new LinkedList<FVhPWaveForm>();
        FVhPWaveForm temp = null;
        $isConditional = false;
        if ( null != wvfrms && false == wvfrms.isEmpty() )
          {
              Iterator<FVhPWaveForm> itr = wvfrms.iterator();
              temp = itr.next();
              while ( itr.hasNext() )
              {
                  FVhPWaveForm curr = itr.next();
                  FVhPExprBase cond = curr.getCondition();
                  // If any of the condition is non-null then it will
                  // be considred as conditional signal assignment
                  if ( null != cond )
                  {
                    $isConditional = true;
                  } 
                  temp.setCondition(cond);
                  $waveformList.add( temp );
                  temp = curr;
                  temp.setCondition(lastCond);
              }
              $waveformList.add( temp );
          }
    }
  :
  waveform 
  {
      firstWvFrm = $waveform.wvfrm;
      if ( null == wvfrms )
      {
          wvfrms = new LinkedList<FVhPWaveForm>();
      }
      wvfrms.add(firstWvFrm);
  }
  ( 
      (WHEN condition ELSE) => cond_wv_bi=conditional_waveforms_bi
      {
          
          if ( null != $cond_wv_bi.elsewvfrms && false == $cond_wv_bi.elsewvfrms.isEmpty() )
          {
              Iterator<FVhPWaveForm> itr = $cond_wv_bi.elsewvfrms.iterator();
              while ( null != wvfrms && itr.hasNext() )
              {
                  wvfrms.add( itr.next() );
              }
          }
      }
  )?
   
  (
      WHEN cond2=condition 
      {
          lastCond = $cond2.expr;
      }
  )?
      
  ;

conditional_waveforms_bi returns [ LinkedList<FVhPWaveForm> elsewvfrms]
  :
  WHEN condition ELSE wv1=waveform 
  {
    if ( null == $elsewvfrms )
      {
          $elsewvfrms = new LinkedList<FVhPWaveForm>();
      }
      FVhPExprBase cond = $condition.expr;
      $wv1.wvfrm.setCondition( cond );
      $elsewvfrms.add( $wv1.wvfrm );
  }
  ( (WHEN condition ELSE) => wvfrm2=conditional_waveforms_bi
  {
    if ( null != $wvfrm2.elsewvfrms && false == $wvfrm2.elsewvfrms.isEmpty() )
      {
          Iterator<FVhPWaveForm> itr = $wvfrm2.elsewvfrms.iterator();
          while ( itr.hasNext() )
          {
              $elsewvfrms.add( itr.next() );
          }
      }
  }
  )?
  ;

//configuration_declaration
//  : CONFIGURATION identifier OF name IS
//    configuration_declarative_part
//    block_configuration
//    END ( CONFIGURATION )? ( identifier )? SEMI
//  ;

configuration_declaration returns[FVhPConfigurationDeclaration du]
  :
  c1=CONFIGURATION nm=identifier OF nm1=name IS 
  {
  	FVhPEntity ent = null;
  	try
  	{
  		ent = (FVhPEntity)FVhPDescription.instance().restoreDesignUnit(workLib, $nm1.tok.getText());
  	}
  	catch (IOException e) 
    {
    	// TODO Auto-generated catch block
    	e.printStackTrace();
    }	
  	$du = new FVhPConfigurationDeclaration( $c1.getLine(), $nm.tok.getText(), ent ); 
  	FVhPDescription.instance().pushScope( $du );
  }
  configuration_declarative_part 
 
  blk=block_configuration
  {
  	$du.setBlockConfiguration($blk.blkcfg);
  }
  END (CONFIGURATION)? (nm2=identifier )? SEMI
  {
    FVhPDescription.instance().popScope();
  }
  ;

configuration_declarative_item
  :
  use_clause
  | attribute_specification
  | group_declaration
  ;

configuration_declarative_part
  :
  (configuration_declarative_item)*
  ;

configuration_item returns[FVhPBase cfg]
  :
  b=block_configuration
  {
  	$cfg = $b.blkcfg;
  }
  | 
  c=component_configuration
  {
  	$cfg = c.compconfig;
  }
  ;

configuration_specification returns[FVhPConfigurationSpecification confspec]
  :
  FOR c=component_specification b=binding_indication SEMI
  {
  	$confspec = new FVhPConfigurationSpecification( $FOR.getLine(),  $b.binding, $c.compspec );
  	FVhPDescription desc = FVhPDescription.instance();
  	desc.addDeclaration( $confspec );
  	currCompSpec = null;
  }
  ;

constant_declaration returns [LinkedList<FVhPConstant> constobjs ]
    @init
    {
        LinkedList<Token> ids = null;
        FVhPSubTypeIndication sind = null;
        FVhPExprBase initval = null;
    }
    @after
    {
        myCurrentSind = null;
    }
  :
  CONSTANT idents=identifier_list{ ids = $idents.ids;} COLON si=subtype_indication{sind = $si.sind; myCurrentSind = $si.sind;}
   (VARASGN expression { initval = $expression.expr;} )? SEMI
   {
       $constobjs  =  new LinkedList<FVhPConstant>();
       if ( null != ids &&  false == ids.isEmpty() )
     {
        Iterator itr = ids.iterator(); 
        while(itr.hasNext()) 
        {
            Token tok = (Token)itr.next(); 
            String constname = tok.getText();
            
            FVhPIdentifier ident = new FVhPIdentifier( tok.getLine(), constname );
            FVhPConstant constobj = new FVhPConstant ( tok.getLine(), ident, sind );
            FVhPDescription desc = FVhPDescription.instance();
            FVhPBase sym = (FVhPBase)desc.getSymbol( constname );
            if ( null != sym && desc.getScope() == sym.getScope() )
            {
                LogMessages.instance().print(3591, LogMessages.MSG_SEVERITY.ERROR, "Found already defined symbol '" + 
                                    constname + "' reference in " +
                                       desc.getFile() +
                                       ":" + tok.getLine() + 
                                       " in file " + sym.getFile() + ":"  + sym.getLine() );
            } 
            desc.insertSymbol( constname, constobj );
            desc.addDeclaration(constobj);
            if ( null != initval )
            {
                initval.setSubtypeIndication( sind );
            }
            constobj.setInitialValue(initval);
            $constobjs.add(constobj);
        }
    }
   }  
  ;

unconstrained_array_definition returns [FVhPTypeDefBase def]
    @init
    {
          LinkedList<FVhPBase> ranges = null;
          boolean isUnconstrained = false;
      }
  :
  ARRAY LPAREN ind1=index_subtype_definition
  {
      ranges = new LinkedList<FVhPBase>();
      if ( null == $ind1.obj )
    {
        LogMessages.instance().print(3590, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+
                                      $ind1.tok.getText() + 
                                      "' in " + FVhPDescription.instance().getFile() +
                                      ":" + $ind1.tok.getLine() );
    }
      FVhPBase range = $ind1.obj;
      ranges.add(range);
      if ( false == isUnconstrained )
      {
        isUnconstrained = $ind1.isUnconstrained;
      }
  }
  (
      COMMA ind2=index_subtype_definition
      {
          if ( null == $ind2.obj )
        {
            LogMessages.instance().print(3589, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+
                                      $ind2.tok.getText() + 
                                      "' in " + FVhPDescription.instance().getFile() +
                                      ":" + $ind2.tok.getLine() );
        }
        FVhPBase range = $ind2.obj;
        if ( null != ranges )
        {
            ranges.add(range);
        }
        if ( false == isUnconstrained )
        {
            isUnconstrained = $ind2.isUnconstrained;
        }
      }
  )* RPAREN OF 
  subtype_indication
  {
      $def = new FVhPArrayType( ranges, $subtype_indication.sind, $subtype_indication.sind.getLine() );
      ((FVhPArrayType)$def).setUnconstrained( isUnconstrained );
  }
  ;

constrained_array_definition returns [FVhPTypeDefBase def]
  :
  ARRAY index_constraint OF subtype_indication
  {
      $def = new FVhPArrayType( $index_constraint.rnglist, $subtype_indication.sind, $subtype_indication.sind.getLine() );
  }
  ;


constraint returns [FVhPConstraint cons]
  :
  r1=range_constraint
  {
          $cons = new FVhPRangeConstraint ( $range_constraint.rng, 
                                        $range_constraint.rng.getLine() );
  }
  | 
  index_constraint
  {
      //LinkedList<FVhPRange> rnglist
      $cons = new FVhPIndexConstraint( $index_constraint.rnglist, $index_constraint.line) ;
  }
  ;


context_clause returns [
                         LinkedList<String> libs, 
                         LinkedList<String> uses, 
                         LinkedList<FVhPLibrary> libObjs,
                         LinkedList<FVhPSelectedName> useClauses
                       ]
  :
  (
      context_item
      {
          if (null == $libs )
          {
              $libObjs = $context_item.libObjs; 
              $libs = $context_item.libs;
          }
          else if ( null != $context_item.libs && false == $context_item.libs.isEmpty() )
          {
              Iterator<String> itr = $context_item.libs.iterator();
              Iterator<FVhPLibrary> libitr = $context_item.libObjs.iterator();
              while ( itr.hasNext() )
              {
                  String libName = itr.next(); 
                  $libs.add( libName );
                  FVhPLibrary libObj = libitr.next(); 
                  $libObjs.add( libObj );
                  FVhPDescription.instance().addSymbol( libName, libObj );
              }
          }
          if ( null == $uses )
          {
              $uses = new LinkedList<String>();
          }
          if ( null != $context_item.uses && false == $context_item.uses.isEmpty() )
          {
              Iterator<String> itr = $context_item.uses.iterator();
              while ( itr.hasNext() )
              {
                  $uses.add( itr.next() );
              }
          }
          
          if ( null == $useClauses )
          {
            $useClauses = new LinkedList<FVhPSelectedName>();
          }
          if ( null != $context_item.useClauses && false == $context_item.useClauses.isEmpty() )
          {
              Iterator<FVhPSelectedName> itr2 = $context_item.useClauses.iterator();
              while ( itr2.hasNext() )
              {
                  $useClauses.add( itr2.next() );
              }
          }
          FVhPDescription.instance().setCurrentUseClauses( $useClauses );
      }
  )*
  ;

context_item returns [ LinkedList<String> libs, 
                       LinkedList<String> uses, 
                       LinkedList<FVhPLibrary> libObjs,
                       LinkedList<FVhPSelectedName> useClauses
                     ]
  :
  library_clause
  {
      libs = $library_clause.libs;
      $libs = libs;
      $libObjs = $library_clause.libObjs;
  }
  | use_clause
  {
      uses = $use_clause.uses;
      $uses = uses;
//      useClauses = $use_clause.useClauses;
      $useClauses = $use_clause.useClauses;
  }
  ;

delay_mechanism
  :
  TRANSPORT
  | (REJECT expression)? INERTIAL
  ;

design_file returns [ FVhPDescription desc ]
  :
  (
      design_unit
      {
          $desc = FVhPDescription.instance();
      }
  )* EOF
  ;


design_unit returns [ FVhPBase du ]
    @init
    {
        FVhPBase.ObjType du_type = FVhPBase.ObjType.UNKNOWN;
        String STDLIB = new String("STD");
        FVhPLibrary lib = new FVhPLibrary(-1, STDLIB);
        FVhPDescription.instance().addSymbol(STDLIB, lib );
    }
  :
  cntxt=context_clause 
  {
      if ( null != $cntxt.uses &&  false == $cntxt.uses.isEmpty() )
      {
          Iterator<String> itr = $cntxt.uses.iterator();
          while ( itr.hasNext() )
          {
              String useclause = itr.next();
              String []temp = useclause.split("\\.");
              try 
            {
                FVhPBase du = FVhPDescription.instance().restoreDesignUnit(temp[0], temp[1]);
//                FVhPDescription.instance().getRoot().addDesignUnit( du );
            } 
            catch (IOException e) 
            {
                // TODO Auto-generated catch block
                e.printStackTrace();
            }
          }
      }
  }
  library_unit
  {
      $du = $library_unit.du;
      du_type = $du.getObjType();
      switch( du_type )
      {
          case ENTITY:
              ((FVhPEntity)$du).setLibraries($context_clause.libObjs);
              ((FVhPEntity)$du).setUseClauses($context_clause.useClauses);
//            ((FVhPEntity)$du).setUseClauses($context_clause.uses);
            break;
        case CONFIGURATIONDECLARATION:
              ((FVhPConfigurationDeclaration)$du).setLibraries($context_clause.libObjs);
              ((FVhPConfigurationDeclaration)$du).setUseClauses($context_clause.useClauses);
//            ((FVhPConfigurationDeclaration)$du).setUseClauses($context_clause.uses);
            break;
            
        case ARCHITECTURE:
              ((FVhPArchitecture)$du).setLibraries($context_clause.libObjs);
//            ((FVhPArchitecture)$du).setUseClauses($context_clause.uses);
            ((FVhPArchitecture)$du).setUseClauses($context_clause.useClauses);
            break;
        case PACKAGE:
              ((FVhPPackage)$du).setLibraries($context_clause.libObjs);
//            ((FVhPPackage)$du).setUseClauses($context_clause.uses);
            ((FVhPPackage)$du).setUseClauses($context_clause.useClauses);
            break;
        case PACKAGEBODY:
              ((FVhPPackageBody)$du).setLibraries($context_clause.libObjs);
//            ((FVhPPackageBody)$du).setUseClauses($context_clause.uses);
            ((FVhPPackageBody)$du).setUseClauses($context_clause.useClauses);
            break;                        
      }
    
    try 
    {
        FVhPDescription.instance().dumpDesignUnit(workLib, $du);
    } 
    catch (IOException e) 
    {
        // TODO Auto-generated catch block
        e.printStackTrace();
    }
    libs = null;
    uses = null;
    FVhPDescription.instance().setCurrentUseClauses( null );
  }  
  ;

designator returns [Token tok]
  :
  identifier
  {
      $tok = $identifier.tok;
  }
  | STRING_LITERAL
  {
      $tok = $STRING_LITERAL;
  }
  ;


direction returns [FVhPRange.Direction dir]
  :
  TO 
    {
              $dir = FVhPRange.Direction.TO;
      }
  | 
  DOWNTO 
      {
              $dir = FVhPRange.Direction.DOWNTO;
      }
  ;

disconnection_specification
  :
  DISCONNECT guarded_signal_specification AFTER expression SEMI
  ;

discrete_range returns [ FVhPBase rng]
  :
  (range) => r1=range
  {
      $rng = $r1.rng;
  }
  |
  s1=subtype_indication
  {
      $rng = $s1.sind;
  }
  ;

element_association returns [FVhPElementAssociation eleassoc]
    @init
    {
        LinkedList<FVhPBase> chcs = null;
    }
  :
  ( (choices ARROW) => choices ARROW { chcs = $choices.choices; })? expression
  {
      $eleassoc = new FVhPElementAssociation( $expression.expr.getLine(), chcs, $expression.expr );
  }
  ;

element_declaration returns [LinkedList<Token> ids, FVhPSubTypeIndication sind]
  :
  identifier_list COLON element_subtype_definition SEMI
  {
      $ids = $identifier_list.ids;
      $sind = $element_subtype_definition.sind;
  }
  ;

element_subtype_definition returns [FVhPSubTypeIndication sind]
  :
  subtype_indication
  {
      $sind = $subtype_indication.sind;
  }
  ;

//entity_aspect
//  : ENTITY name ( LPAREN identifier RPAREN )?
//  | CONFIGURATION name
//  | OPEN
//  ;

entity_aspect returns[FVhPSelectedName prim, Token tok, Token sec, FVhPBase.ObjType duType, boolean isOpen]
  :
  ENTITY nm1=name
  {
  	$duType = FVhPBase.ObjType.ENTITY;
  	$prim = (FVhPSelectedName)$nm1.selectedName;
  	$tok = $nm1.tok;
  } 
  (LPAREN id=identifier
  {
  	$sec = $id.tok;
  } 
  RPAREN)?
  | 
  CONFIGURATION nm2=name
  {
//  	$duType = FVhPBase.ObjType.CONFIGURATION;
  	$prim = (FVhPSelectedName)$nm2.selectedName;
  	LogMessages.instance().print(3588, LogMessages.MSG_SEVERITY.WARNING, "configuration in entity_aspect is not supported yet ...");
  }
  | 
  OPEN
  {
  	$isOpen = true;
  	LogMessages.instance().print(3587, LogMessages.MSG_SEVERITY.ERROR, "OPEN in entity_aspect is not supported yet ...");
  }
  ;

entity_class
  :
  ENTITY
  | ARCHITECTURE
  | CONFIGURATION
  | PROCEDURE
  | FUNCTION
  | PACKAGE
  | TYPE
  | SUBTYPE
  | CONSTANT
  | SIGNAL
  | VARIABLE
  | COMPONENT
  | LABEL
  | LITERAL
  | UNITS
  | GROUP
  | FILE
  ;

entity_class_entry
  :
  entity_class (BOX)?
  ;

entity_class_entry_list
  :
  entity_class_entry (COMMA entity_class_entry)*
  ;

//entity_declaration
//  : ENTITY identifier IS entity_header
//    entity_declarative_part
//    ( BEGIN entity_statement_part )?
//    END ( ENTITY )? ( identifier )? SEMI
//  ;

entity_declaration returns [FVhPEntity entity]
  @init 
  { 
      String name = null;
      LinkedList<FVhPPort> ports = null;
      LinkedList<FVhPGeneric> generics = null;
  }
  :
  ENTITY id=identifier 
          {
              name = $id.tok.getText();
              FVhPDescription desc = FVhPDescription.instance();
              FVhPBase sym = desc.getSymbol(name);
              if (null != sym && ( FVhPBase.ObjType.ENTITY == sym.getObjType() || FVhPBase.ObjType.PACKAGE == sym.getObjType() || FVhPBase.ObjType.CONFIGURATIONDECLARATION == sym.getObjType() ) )
              {
              	LogMessages.instance().print(3586, LogMessages.MSG_SEVERITY.ERROR, 
              	                     "Symbol '"+
                                      name + 
                                      "' is already defined at " + FVhPDescription.instance().getFile() +
                                      ":" + $id.tok.getLine() + " in the scope " + sym.getScope());
                     
              }
              $entity = new FVhPEntity( name,  $id.tok.getLine() );
              desc.insertSymbol( name, $entity);
              desc.pushScope($entity);
          } IS 
          entity_header
          {
              ports = $entity_header.ports;
            generics = $entity_header.generics;
          } 
          entity_declarative_part 
  (BEGIN entity_statement_part)? 
  END (ENTITY)? (identifier)? SEMI
  {         
        $entity.setGenerics( generics );
        $entity.setPorts( ports );
//        $entity.setLibraries(libs);
//        $entity.setUseClauses(uses);
        FVhPDescription desc = FVhPDescription.instance();
        desc.popScope();
        FVhPDescription.instance().getRoot().addDesignUnit($entity);
//        $entity.setLibraries(libs);
//        $entity.setUseClauses(uses);
        libs = null;
        uses = null;
  }
  ;


entity_declarative_item
  :
  (subprogram_declaration) => subprogram_declaration
  | subprogram_body
  | type_declaration
  | subtype_declaration
  | constant_declaration
  | signal_declaration
  | variable_declaration
  | file_declaration
  | alias_declaration
  | attribute_declaration
  | attribute_specification
  | disconnection_specification
  | use_clause
  | group_template_declaration
  | group_declaration
  ;

entity_declarative_part
  :
  (entity_declarative_item)*
  ;

entity_designator
  :
  entity_tag (signature)?
  ;

entity_header  returns [LinkedList<FVhPPort> ports,
                        LinkedList<FVhPGeneric> generics]
  :
  (generic_clause 
      {
              $generics = $generic_clause.generics;
      }
  )? 
  (port_clause
      {
              $ports = $port_clause.ports;
      }
  )?
  ;


entity_name_list
  :
  entity_designator (COMMA entity_designator)*
  | OTHERS
  | ALL
  ;

entity_specification
  :
  entity_name_list COLON entity_class
  ;

entity_statement
  :
  (
    (label_colon)? (POSTPONED)? ASSERT
  )
    => concurrent_assertion_statement
  |
  (
    (label_colon)? (POSTPONED)? PROCESS
  )
    => process_statement
  | concurrent_procedure_call_statement
  ;

entity_statement_part
  :
  (entity_statement)*
  ;

entity_tag
  :
  identifier
  | CHARACTER_LITERAL
  | STRING_LITERAL
  ;

enumeration_literal returns [Token tok, boolean isCharLit]
  :
  identifier
  {
      $tok = $identifier.tok;
      $isCharLit = false;
  }
  | CHARACTER_LITERAL
  {
      $tok = $CHARACTER_LITERAL;
      $isCharLit = true;
  }
  ;

enumeration_type_definition returns [Integer line, LinkedList<FVhPEnumLiteral> literals, boolean isCharLit]
  :
  LPAREN en1=enumeration_literal
  {
    $line = $en1.tok.getLine();
    $isCharLit = $en1.isCharLit;
      $literals = new LinkedList<FVhPEnumLiteral>();
      FVhPEnumLiteral lit = new FVhPEnumLiteral( $line, $en1.tok.getText(), $isCharLit );
      $literals.add( lit );
      FVhPDescription.instance().addSymbol( $en1.tok.getText(), lit );
  } 
  (
      COMMA en2=enumeration_literal
      {
          $isCharLit = $en2.isCharLit;
          $line = $en2.tok.getLine();
          FVhPEnumLiteral lit = new FVhPEnumLiteral( $line, $en2.tok.getText(), $isCharLit );
          $literals.add( lit );
          FVhPDescription.instance().addSymbol( $en2.tok.getText(), lit );
      }
  )* RPAREN
  ;

exit_statement
  :
  (label_colon)? EXIT (identifier)? (WHEN condition)? SEMI
  ;

// NOTE that NAND/NOR are in (...)* now (used to be in (...)?).
// (21.1.2004, e.f.)


expression returns [FVhPExprBase expr]
    :
  r1=relation
  {
  	$expr = $r1.expr;
  } 
  (options {greedy=true;}: 
      logical_operator r2=relation
      {

              $expr = new FVhPExprBinary( $r2.expr.getLine(), $expr, $r2.expr, $logical_operator.op );
      }
  )*
  ;

factor returns [FVhPExprBase expr]
    @init
    {
        FVhPExprBase lhs = null;
        FVhPExprBase rhs = null;
        FVhPExprBinary.OpType op = FVhPExprBinary.OpType.UNKNOWN;
    }
  :
  (
  p1=primary
  {
      lhs = $p1.expr;
  }
   (options {greedy=true;}: DOUBLESTAR 
           p2=primary
           {
                   rhs = $p2.expr;
           }
   )?
  )
  {
          if ( null != rhs )
          {
              $expr = new FVhPExprBinary ( ( null == lhs ) ? 0 : lhs.getLine(), lhs, rhs, op );
          }
          else
          {
              $expr = lhs;
          }
  }
  | ABS p3=primary
  {
          $expr = new FVhPExprUnary($p3.expr.getLine(), $p3.expr, FVhPExprUnary.OpType.ABS ); 
  }
  | NOT p4=primary
  {
            $expr = new FVhPExprUnary($p4.expr.getLine(), $p4.expr, FVhPExprUnary.OpType.NOT ); 
  }
  ;

file_declaration returns[FVhPFileDecl filedecl]
	@init
	{
		FVhPExprBase openkind = null;
		FVhPExprBase logicalname = null;
	}
  :
  FILE identifier_list COLON subtype_indication 
  (
  file_open_information
  {
  	logicalname = $file_open_information.logicalname;
  	openkind = $file_open_information.openKind;
  }
  )? SEMI
  {
  	LinkedList<Token> ids = $identifier_list.ids;
     if ( null != ids && false == ids.isEmpty() )
     {
        Iterator itr = ids.iterator(); 
        while(itr.hasNext()) 
        {
            Token tok = (Token)itr.next(); 
            String name = tok.getText();
            FVhPDescription desc = FVhPDescription.instance();
            FVhPBase sym = (FVhPBase)desc.getSymbol( name );
            if ( null != sym && desc.getScope() == sym.getScope() )
            {
                LogMessages.instance().print(3585, LogMessages.MSG_SEVERITY.ERROR, 
                                    "Found already defined symbol '" + 
                                    name + "' reference in " +
                                       desc.getFile() +
                                       ":" + tok.getLine() + 
                                       " in file " + sym.getFile() + ":"  + sym.getLine());
            } 
            $filedecl = new FVhPFileDecl($FILE.getLine(), name, $subtype_indication.sind); 
            $filedecl.setOpenKind( openkind );
            $filedecl.setLogicalName( logicalname ); 
            desc.insertSymbol( name, $filedecl );
            desc.addDeclaration( $filedecl );
        } 
    } 
  }
  ;

file_logical_name returns[FVhPExprBase logicalname]
  :
  expression
  {
  	$logicalname = $expression.expr;
  }
  ;

file_open_information returns[FVhPExprBase openKind, FVhPExprBase logicalname]
  :
  (OPEN expression
  {
  	$openKind = $expression.expr;
  }
  )? IS 
  file_logical_name
  {
  	$logicalname = $file_logical_name.logicalname;
  }
  ;

file_type_definition returns[FVhPTypeDefBase typedef]
  :
  FILE OF name
  {
  	FVhPDescription desc = FVhPDescription.instance();
  	FVhPBase sind = desc.getSymbol( $name.tok.getText() );
  	if ( null == sind )
  	{
  		LogMessages.instance().print(3584, LogMessages.MSG_SEVERITY.ERROR, "Could not find definition of the subtype indication " +
					                  $name.tok.getText() + " used at " + desc.getFile() + ":" + $name.tok.getLine() + " ..." );
  	}
  	else
  	{
  		$typedef = new FVhPFileType( $FILE.getLine(), sind );
  	}
  }
  ;

formal_parameter_list returns [LinkedList<FVhPInterfaceElement> intf_ele]
  :
  interface_list
  {
      $intf_ele = $interface_list.intf_ele ;
  }
  ;

formal_part returns[Token formal]
  :
  nm1=name
  {
  	$formal = $nm1.tok;
  }
  (LPAREN 
  nm2=name
  {
  	LogMessages.instance().print(3583, LogMessages.MSG_SEVERITY.ERROR, "Unsupported port mapping found at " +
  	                              $LPAREN.getLine() );
  } 
  RPAREN)?
  ;

//function_call
//  :
//  name (LPAREN actual_parameter_part RPAREN)?
//  ;

function_call returns [FVhPExprBase expr]
    @init
    {
        boolean isFuncCall = false;
        FVhPBase func = null;
        FVhPDescription desc = FVhPDescription.instance();
    }
  :
  name
  {
  	func =  desc.getSymbol( $name.tok.getText() );
  	currSubprog = func;
  } 
  (LPAREN actual_parameter_part { isFuncCall = true; } RPAREN)?
  {
  	  
  	  
  	  FVhPBase.ObjType objtype = ( null != func ) ? func.getObjType()  : FVhPBase.ObjType.UNKNOWN ;
  	  if ( ( false == $name.isBitString ) && ( ( true == isFuncCall || null == func ) || 
  	       ( FVhPBase.ObjType.SUBPROGRAM_DECL == objtype ) || 
  	       ( FVhPBase.ObjType.SUBPROGRAM_BODY == objtype ) ) )
  	  {
  	  	isFuncCall = true; 
  	  	$expr = new FVhPExprFunctionCall( $name.tok.getLine(), $name.tok.getText() );
  	  	((FVhPExprFunctionCall)$expr).setSubprogramDefinition( func );
  	  	if ( null != $actual_parameter_part.paramlist )
  	  	{
  	  		((FVhPExprFunctionCall)$expr).setArguments($actual_parameter_part.paramlist);
  	  	}
  	  	else if ( null != $name.bitsel )
  	  	{
  	  		 LinkedList<FVhPExprBase> args = ((FVhPBitSelect) $name.bitsel).getIndexList();
  	  		 ((FVhPExprFunctionCall)$expr).setArguments( args );
  	  	}
  	  }
      else if ( false == $name.isBitString )
      {
        // Ensure that this symbol is already defined and
        // also ensure that this is not function 
        desc = FVhPDescription.instance();
  	    FVhPBase sym = $name.obj;
  	   
  	    if ( null == sym )
  	    {
      		sym = FVhPDescription.instance().getSymbol($name.tok.getText());
        }        
        if ( null == sym )
        {
            LogMessages.instance().print(3582, LogMessages.MSG_SEVERITY.ERROR, "Found reference to undefined symbol '"+ $name.tok.getText() + 
                                       "' in " +
                                       desc.getFile() +
                                       ":" + $name.tok.getLine() );
        }
         
        if ( null != $name.attribName )
        {
            $expr = new FVhPExprAttribute($name.tok.getLine(), sym, $name.attribName );
        }
        else if ( null != $name.rangelist && false == $name.rangelist.isEmpty() )
        {
	      if ( null != $name.declaredObj && FVhPBase.ObjType.SIMPLENAME != $name.declaredObj.getObjType() )
	      {
	       sym = $name.declaredObj;
	      }
          $expr = new FVhPExprPartSelect($name.tok.getLine(), sym, $name.rangelist);
        }
        else if ( null != $name.bitsel )
        {
            if ( sym instanceof FVhPSimpleName )
            {
//                FVhPDescription.instance().addSymbol( ((FVhPSimpleName)sym).getName(), ((FVhPBitSelect)$name.bitsel).getParent() );
                sym = ((FVhPBitSelect)$name.bitsel).getParent();
            }
        	$expr = new FVhPExprBitSelect( $name.tok.getLine(), sym, ((FVhPBitSelect)$name.bitsel).getIndexList() );
        }
        else if ( null != $name.aggregate )
        {
        	FVhPExprBase aggr = new FVhPExprAggregate( $name.aggregate.getLine(), (FVhPAggregate)$name.aggregate);
        	$expr = new FVhPExprQualified( $name.tok.getLine(), sym, aggr );
        }
        else
        {
        	FVhPBase currSym = sym;
        	
        	if ( null != sym && FVhPBase.ObjType.SIMPLENAME == sym.getObjType() )
        	{
        		sym = FVhPDescription.instance().getSymbol( ((FVhPSimpleName)sym).getName() );
        		sym = ( null != sym ) ?  sym : currSym;
        	}
        	
            $expr = new FVhPExprIdentifier( $name.tok.getLine(), sym );
        }          
      }
      else if ( true == $name.isBitString )
      { 
          String str = $name.tok.getText();
          String valWithoutQuotes = str.substring(1, str.length()-1);
          if ( valWithoutQuotes.matches("[^01xXzZ_]+") )
          {
            $expr = new FVhPExprBitString($name.tok.getLine(), valWithoutQuotes);
          }
          else
          {
            $expr = new FVhPExprBitString($name.tok.getLine(), valWithoutQuotes);
          }
      }
      else
      {
          LogMessages.instance().print(3581, LogMessages.MSG_SEVERITY.ERROR, "Unsupported token found at line " + $name.tok.getLine() + " ...");
      }
      currSubprog = null;
  }
  ;

generate_statement returns[FVhPBase stmt]
  :
  label_colon generation_scheme g1=GENERATE
  {
  	String label = $label_colon.tok.getText();
  	$stmt = new FVhPGenerateStatement( $g1.getLine(), 
								       $generation_scheme.scheme,
			                           $generation_scheme.param,
			                           $generation_scheme.cond);
	((FVhPGenerateStatement)$stmt).setLabel(label);
	if ( null != $generation_scheme.param )
	{	
	   $generation_scheme.param.setScope( $stmt );
	}
	if ( null != $generation_scheme.forindex )
	{
	   $generation_scheme.forindex.setScope ( $stmt );
	}
	FVhPDescription.instance().addStatement($stmt);
	FVhPDescription.instance().pushScope($stmt);
  } 
  ( (block_declarative_item)* BEGIN)? (architecture_statement)* 
  END GENERATE (identifier)? SEMI
  {
  	FVhPDescription.instance().popScope();
  }
  ;

generation_scheme returns[Integer line, 
                          FVhPGenerateStatement.SchemeType scheme,
                          FVhPLoopParameter param,
                          FVhPForIndex forindex,
                          FVhPExprBase cond
                          ]
  :
  FOR parameter_specification
  {
  	$scheme = FVhPGenerateStatement.SchemeType.FOR;
  	$line = $FOR.getLine();
  	$param = $parameter_specification.param;
  	$forindex = $parameter_specification.forindex;
  	$cond = null;
  }
  | IF condition
  {
  	$scheme = FVhPGenerateStatement.SchemeType.IF;
  	$line = $IF.getLine();
  	$cond = $condition.expr;
  	$param = null;
  }
  ;

generic_clause returns [LinkedList<FVhPGeneric> generics]
  :
  GENERIC LPAREN 
  generic_list
  {
          $generics = new LinkedList<FVhPGeneric>();
          LinkedList<FVhPInterfaceElement> intfs = $generic_list.intf_ele;
            if ( intfs != null && false == intfs.isEmpty() )
        {
            Iterator itr = intfs.iterator(); 
            while(itr.hasNext()) 
            {
                FVhPInterfaceElement intf = (FVhPInterfaceElement)itr.next(); 
                FVhPGeneric gen = new FVhPGeneric(intf);
                $generics.add(gen);
                String genName = intf.getIdentifier().getName();
                FVhPDescription desc = FVhPDescription.instance(); 
                desc.insertSymbol( genName, gen );
            } 
        }
  } 
  RPAREN SEMI
  ;

generic_list returns [LinkedList<FVhPInterfaceElement> intf_ele]
  :
  interface_list
  {
          $intf_ele =  $interface_list.intf_ele;
  }
  ;


generic_map_aspect returns[LinkedList<FVhPGenericMap> genericmaps]
  :
  GENERIC MAP LPAREN  association_list RPAREN
  {
  	$genericmaps = $association_list.genericmaps;
  	if ( null == $genericmaps && null != $association_list.actuals && 
  	     false == $association_list.actuals.isEmpty() )
  	{
  		LinkedList<FVhPGeneric> generics = getCurrentInstanceMasterGenericList();
  		
  		Iterator<FVhPGeneric> itr2 = null;
  		if ( null != generics && false == generics.isEmpty() )
  		{
  		    itr2 = generics.iterator();
  		} 
  		$genericmaps = new LinkedList<FVhPGenericMap>();
		Iterator<FVhPExprBase> itr = $association_list.actuals.iterator();
		while ( itr.hasNext() )
		{
			FVhPExprBase expr = itr.next();
			FVhPGeneric  formal = ( null != itr2 ) ? itr2.next() : null ;
			FVhPGenericMap map = new FVhPGenericMap( $LPAREN.getLine(), formal, expr );
			$genericmaps.add( map );
		}
  	}
  }
  ;

group_constituent
  :
  name
  | CHARACTER_LITERAL
  ;

group_constituent_list
  :
  group_constituent (COMMA group_constituent)*
  ;

group_declaration
  :
  GROUP label_colon name LPAREN group_constituent_list RPAREN SEMI
  ;

group_template_declaration
  :
  GROUP identifier IS LPAREN entity_class_entry_list RPAREN SEMI
  ;

guarded_signal_specification
  :
  signal_list COLON name
  ;

//identifier
//  : BASIC_IDENTIFIER
//  | EXTENDED_IDENTIFIER
//  ;

identifier returns [Token tok]
  :
  id=BASIC_IDENTIFIER 
  {
      $tok = $id;
  }
  | eid=
  EXTENDED_IDENTIFIER 
  { 
      $tok = $eid;
  }
  ;

identifier_list returns [LinkedList<Token> ids]
    @init { LinkedList<Token> idlist = null; }
  :
  id1=identifier
  {
      if ( null == idlist )
      {
          idlist = new LinkedList<Token>();
      }
      idlist.add($id1.tok);
  } 
  (COMMA 
      id2=identifier
      {
          if ( null == idlist )
          {
              idlist = new LinkedList<Token>();
          }
          idlist.add($id2.tok);
      }
  )*
  {
      $ids = idlist;
  }
  ;

if_statement
    @init
    {
        String label = null;
        FVhPSeqIfStmt ifstmt = null;
        FVhPDescription desc = FVhPDescription.instance();
    }
  :
  (label_colon{ label = $label_colon.tok.getText();})? IF c1=condition THEN
  {
      ifstmt =  new FVhPSeqIfStmt( $c1.expr.getLine(), $c1.expr );
      ifstmt.setLabel(label);
      desc.addStatement(ifstmt);
      desc.pushScope(ifstmt);
  } 
  sequence_of_statements 
  (
      ELSIF c2=condition THEN 
      {
          FVhPSeqElsif elsif = new FVhPSeqElsif( $c2.expr.getLine(), $c2.expr );
          if ( null != ifstmt )
          {
            ifstmt.addElsifStmt(elsif);
          }
          desc.pushScope(elsif);
      }
      sequence_of_statements
      {
          desc.popScope();
      }
  )* 
  (
      ELSE
      {
          FVhPSeqElse elseblk = new FVhPSeqElse( $ELSE.getLine());
          if ( null != ifstmt )
          {
            ifstmt.setElseStmtBlock(elseblk);
          }
          desc.pushScope(elseblk);
      } 
      sequence_of_statements
      {
          desc.popScope();
      }
  )? 
  END IF (identifier)? SEMI
  {
      desc.popScope();
  }
  ;

index_constraint returns [LinkedList<FVhPBase> rnglist, Integer line]
  :
  LPAREN dr1=discrete_range
  {
      $rnglist = new LinkedList<FVhPBase>();
      $rnglist.add( $dr1.rng );
      $line = $dr1.rng.getLine();
  } 
  (
  COMMA 
      dr2=discrete_range
      {
          $rnglist.add( $dr2.rng );
      }
  )* RPAREN
  ;

index_specification returns [FVhPBase rng, FVhPExprBase expr]
  :
  (discrete_range) => dr=discrete_range
  {
      $rng = $dr.rng;
  }
  | expression
  {
  	$expr = $expression.expr;
  }
  ;

index_subtype_definition returns [
                                   Token tok, 
                                   FVhPBase obj,
                                   boolean isUnconstrained
                                 ]
  :
  nm=name RANGETOK BOX
  {
      $tok = $nm.tok;
      $obj =  $nm.obj;
      $isUnconstrained = true;
  }
  |
  nm=name r=range_constraint
  {
    $tok = $nm.tok;
    $isUnconstrained = false;
    LinkedList<FVhPBase> ranges = new LinkedList<FVhPBase>();
    ranges.add( $r.rng );
    String typeName = getNameString( $nm.obj );
    FVhPBase typemark = FVhPDescription.instance().getSymbol( typeName );
    $obj = new FVhPSubTypeIndication( $nm.tok.getLine(), typemark, ranges);
  }
  ;

//instantiated_unit
//  : ( COMPONENT )? name
//  | ENTITY name ( LPAREN identifier RPAREN )?
//  | CONFIGURATION name
//  ;

instantiated_unit returns [FVhPBase master, String libName, FVhPArchitecture arch]
	@init
	{
	   String archname = null;
	}
  :
  (COMPONENT)? nm1=name
  {
  	$master = FVhPDescription.instance().getSymbol($nm1.tok.getText());
  	if ( null == $master )
    {
        // Create dummy component in order  to support mixed Language designs
        $master = new FVhPComponentDecl( $nm1.tok.getLine(), $nm1.tok.getText() );
    }
  }
  | ENTITY nm2=name (LPAREN identifier { archname = $identifier.tok.getText(); }  RPAREN)?
  {
  	$master = null;
  	if ( null == $master )
  	{
  		// Load this entity
  		String lib = null;
  		String entName = null;
  		
  		if ( null != $nm2.selectedName )
  		{
  			LinkedList<FVhPBase> eleList = ((FVhPSelectedName)$nm2.selectedName).getElementList();
  			Iterator<FVhPBase> itr = eleList.iterator();
  			while ( itr.hasNext() )
  			{
  				if ( null == lib && null != entName )
  				{
  					lib = entName;
  					entName = null;
  				}
  				FVhPBase item = itr.next();
  				if ( null != item )
  				{
  					entName = item.print();
  				}		
  			}
  			if ( null == lib )
  			{
  				lib = workLib;
  			}
  		}
  		else
  		{
  			lib = workLib;
  			entName = $nm2.tok.getText();
  		}
  		try 
        {
            FVhPBase du = FVhPDescription.instance().restoreDesignUnit(lib,entName);
            $master = du;
            if ( null != archname )
            {
                $arch = (FVhPArchitecture)FVhPDescription.instance().restoreDesignUnit(lib,entName+":"+archname);
            }
        }
        catch (IOException e) 
        {
        	e.printStackTrace();
        }
        $libName = lib;
  	}
  }
  | CONFIGURATION nm3=name
  {
    
  	$master = FVhPDescription.instance().getSymbol($nm3.tok.getText());
  	String confName = null;
  	if ( null == $master )
    {
        String lib = null;        
        if ( null != $nm3.selectedName )
        {
            LinkedList<FVhPBase> eleList = ((FVhPSelectedName)$nm3.selectedName).getElementList();
            Iterator<FVhPBase> itr = eleList.iterator();
            while ( itr.hasNext() )
            {
                if ( null == lib && null != confName )
                {
                    lib = confName;
                    confName = null;
                }
                FVhPBase item = itr.next();
                if ( null != item )
                {
                    confName = item.print();
                }       
            }
            if ( null == lib )
            {
                lib = workLib;
            }
        }
        else
        {
            lib = workLib;
            confName = $nm3.tok.getText();
        }
        try 
        {
            FVhPBase du = FVhPDescription.instance().restoreDesignUnit(lib,confName);
            $master = du;
        }
        catch (IOException e) 
        {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
        $libName = lib;
    }
  }
  ;

instantiation_list returns[LinkedList<String> instNameList, boolean isOthers, boolean isAll]
  :
  id1=identifier
  {
  	$instNameList = new LinkedList<String>();
  	$instNameList.add( $id1.tok.getText() );
  	$isAll = false;
  	$isOthers = false;
  } 
  (COMMA id2=identifier
  {
  	$instNameList.add( $id2.tok.getText() );
  }
  )*
  | OTHERS
  {
  	$isOthers = true;
  	$isAll = false;
  }
  | ALL
  {
  	$isAll = true;
  	$isOthers = false;
  }
  ;


// This is the rule for entity generic and port declaration.
// It seems that the parse flow treats both generics and ports as
// constant interface for the entity.
interface_constant_declaration returns [LinkedList<FVhPInterfaceElement> cons]
    @init
    {
        LinkedList<Token> ids = null;
        FVhPSubTypeIndication sind = null;
        FVhPExprBase initval = null;
    }
  :
  (CONSTANT)? identifier_list {ids = $identifier_list.ids; } 
  COLON (IN)? subtype_indication { sind = $subtype_indication.sind; } 
  (VARASGN expression { initval = $expression.expr ;})?
  {
     $cons  =  new LinkedList<FVhPInterfaceElement>();
     if ( null != ids && false == ids.isEmpty() )
     {
        Iterator itr = ids.iterator(); 
        while(itr.hasNext()) 
        {
            Token tok = (Token)itr.next(); 
            String genName = tok.getText();
            FVhPIdentifier ident = new FVhPIdentifier( tok.getLine(), genName );            
            FVhPInterfaceElement ele = new FVhPInterfaceElement ( tok.getLine() , ident, sind);
            ele.setStorageClass(FVhPInterfaceElement.ObjectClass.CONSTANT);
            FVhPDescription desc = FVhPDescription.instance();
            FVhPBase sym = (FVhPBase)desc.getSymbol( genName );
            if ( null != sym && desc.getScope() == sym.getScope() )
            {
                LogMessages.instance().print(3580, LogMessages.MSG_SEVERITY.ERROR,
                                    "Found already defined symbol '" + 
                                    genName + "' reference in " +
                                       desc.getFile() +
                                       ":" + tok.getLine() + 
                                       " in file " + sym.getFile() + ":"  + sym.getLine());
            } 
            desc.insertSymbol( genName, ele );
            if ( null != initval )
            {
                initval.setSubtypeIndication( sind );
            }
            ele.setInitialValue(initval);
            $cons.add(ele);
        } 
    } 
  }
  ;


interface_declaration returns [LinkedList<FVhPInterfaceElement> intf_ele]
 @init 
 { 
     LinkedList<FVhPInterfaceElement> elements = null;
 }
  :
  ((con1=interface_constant_declaration
      {              
              $intf_ele = $con1.cons;
      }
      ) => con2=interface_constant_declaration
      {       
              $intf_ele = $con2.cons;  
      }
      )
  | 
  ((ifsig=interface_signal_declaration 
      {      
              $intf_ele = $ifsig.sig;
      } 
      ) => ifs=interface_signal_declaration
      {      
              $intf_ele = $ifs.sig;
      })   
  | interface_variable_declaration
  {
          $intf_ele = $interface_variable_declaration.var;
  }   
  | interface_file_declaration
  {  
          $intf_ele = null;
  }
  ;


interface_element returns [LinkedList<FVhPInterfaceElement> intf_ele]
  :
  interface_declaration
  {
          $intf_ele = $interface_declaration.intf_ele;
  }
  ;

interface_file_declaration
  :
  FILE identifier_list COLON subtype_indication
  ;


interface_list returns [LinkedList<FVhPInterfaceElement> intf_ele]                
  :
  if1=interface_element
  {
      if ( null == $intf_ele )
      { 
          $intf_ele = $if1.intf_ele;
      }
      else
      {
          if ( null != $if1.intf_ele && false == $if1.intf_ele.isEmpty() )
        {
            Iterator itr = $if1.intf_ele.iterator(); 
            while(itr.hasNext()) 
            {
                FVhPInterfaceElement ele = (FVhPInterfaceElement)itr.next(); 
                $intf_ele.add(ele);
            } 
        }
      }
  } 
  (SEMI 
      if2=interface_element
      {
          if ( null == $intf_ele )
          { 
              $intf_ele = $if2.intf_ele;
          }
          else
          {
              if ( null != $if2.intf_ele && false == $if2.intf_ele.isEmpty() )
            {
                Iterator itr = $if2.intf_ele.iterator(); 
                while(itr.hasNext()) 
                {
                    FVhPInterfaceElement ele = (FVhPInterfaceElement)itr.next(); 
                    $intf_ele.add(ele);
                } 
            }
          }
      } 
  )*
  ;

interface_signal_declaration returns [LinkedList<FVhPInterfaceElement> sig]
@init
    {
        LinkedList<Token> ids = null;
        FVhPSubTypeIndication sind = null;
        FVhPExprBase initval = null;
        FVhPPort.Direction dir = FVhPPort.Direction.UNKNOWN;
    }
    
  :
   ((SIGNAL)? identifier_list {ids = $identifier_list.ids; } 
   COLON (mode { dir = $mode.dir; })? 
   subtype_indication { sind = $subtype_indication.sind;} 
   (BUS)? 
   (VARASGN expression { initval = $expression.expr; })?)
   {
       $sig  =  new LinkedList<FVhPInterfaceElement>();
       if ( null != ids &&  false == ids.isEmpty() )
     {
        Iterator itr = ids.iterator(); 
        while(itr.hasNext()) 
        {
            Token tok = (Token)itr.next(); 
            String portName = tok.getText();
            
            FVhPIdentifier ident = new FVhPIdentifier( tok.getLine(), portName );
            FVhPInterfaceElement ele = new FVhPInterfaceElement ( tok.getLine(), ident, sind );
            ele.setStorageClass(FVhPInterfaceElement.ObjectClass.SIGNAL);
            ele.setDirection(dir);
            FVhPDescription desc = FVhPDescription.instance();
            FVhPBase sym = (FVhPBase)desc.getSymbol( portName );
            if ( null != sym && desc.getScope() == sym.getScope()  )
            {
                LogMessages.instance().print(3579, LogMessages.MSG_SEVERITY.ERROR,
                                    "Found already defined symbol '" + 
                                    portName + "' reference in " +
                                       desc.getFile() +
                                       ":" + tok.getLine() +
                                       " in file " + sym.getFile() + ":"  + sym.getLine() );
            } 
            desc.insertSymbol( portName, ele );
            if ( null != initval )
            {
                initval.setSubtypeIndication( sind );
            }
            ele.setInitialValue(initval);
            $sig.add(ele);
        } 
    } 
  }
  ;


interface_variable_declaration returns [LinkedList<FVhPInterfaceElement> var]
@init
    {
        LinkedList<Token> ids = null;
        FVhPSubTypeIndication sind = null;
        FVhPExprBase initval = null;
        FVhPPort.Direction dir = FVhPPort.Direction.UNKNOWN;
    }    
  :
   ((VARIABLE)? identifier_list {ids = $identifier_list.ids; } 
   COLON (mode { dir = $mode.dir;})? 
   subtype_indication { sind = $subtype_indication.sind;} 
   (VARASGN expression { initval = $expression.expr; })?)
    {
       $var  =  new LinkedList<FVhPInterfaceElement>();
       if ( null != ids &&  false == ids.isEmpty() )
     {
        Iterator itr = ids.iterator(); 
        while(itr.hasNext()) 
        {
            Token tok = (Token)itr.next(); 
            String varName = tok.getText();
            
            FVhPIdentifier ident = new FVhPIdentifier( tok.getLine(), varName );
            FVhPInterfaceElement ele = new FVhPInterfaceElement ( tok.getLine(), ident, sind );
            ele.setDirection(dir);
            ele.setStorageClass(FVhPInterfaceElement.ObjectClass.VARIABLE);
            FVhPDescription desc = FVhPDescription.instance();
            FVhPBase sym = (FVhPBase)desc.getSymbol( varName );
            if ( null != sym && desc.getScope() == sym.getScope() )
            {
                LogMessages.instance().print(3578, LogMessages.MSG_SEVERITY.ERROR,
                                    "Found already defined symbol '" + 
                                    varName + "' reference in " +
                                       desc.getFile() +
                                       ":" + tok.getLine() +
                                       " in file " + sym.getFile() + ":"  + sym.getLine() );
            } 
            desc.insertSymbol( varName, ele );
            if ( null != initval )
            {
                initval.setSubtypeIndication( sind );
            }
            ele.setInitialValue(initval);
            $var.add(ele);
        } 
    } 
  }
  ;



iteration_scheme returns [ FVhPLoopParameter param, 
                           FVhPExprBase cond, 
                           FVhPSeqLoopStatment.LoopType type ]
  :
  WHILE condition
  {
      $cond = $condition.expr;
      $type = FVhPSeqLoopStatment.LoopType.WHILE;
  }
  | 
  FOR parameter_specification
  {
      $param = $parameter_specification.param;      
      $type = FVhPSeqLoopStatment.LoopType.FOR;
  }
  ;

label_colon returns [Token tok ]
  :
  identifier COLON
  {
      $tok = $identifier.tok;
  }
  ;

library_clause returns [ LinkedList<String> libs,  LinkedList<FVhPLibrary> libObjs]
  :
  LIBRARY logical_name_list SEMI
  {
      libs = $logical_name_list.libs;
      if ( null != libs && false == libs.isEmpty() )
      {
          $libObjs = new LinkedList<FVhPLibrary>();
        Iterator<String> itr = libs.iterator();
        while(itr.hasNext())
        {
            String lbname = itr.next();
            FVhPLibrary lb = new FVhPLibrary($logical_name_list.line, lbname );
            FVhPDescription.instance().addSymbol(lbname, lb);
            $libObjs.add( lb );
        }          
      }
      $libs = libs;
  }
  ;


library_unit returns [FVhPBase du]
  :
  (
    ARCHITECTURE
    | PACKAGE BODY
  )
    => secondary_unit
  {
      $du = $secondary_unit.du;
  }  
  | 
  primary_unit
  {
      $du = $primary_unit.du;
  }
  ;

literal returns [FVhPExprBase expr]
  :
  NULLTOK 
  {
      $expr = new FVhPExprNull( $NULLTOK.getLine() );
  }
  | BIT_STRING_LITERAL
  {
  	String text = $BIT_STRING_LITERAL.getText();
  	String withoutQuote = text.substring( 2, text.length() - 1 ); 
  	$expr = new FVhPExprBitString($BIT_STRING_LITERAL.getLine(), withoutQuote );
  	((FVhPExprBitString)$expr).setBase( text.charAt(0) );
  }
  | (DBLQUOTE) => STRING_LITERAL
  {
          LogMessages.instance().print(3577, LogMessages.MSG_SEVERITY.WARNING, "STRING_LITERAL expression is not supported yet ...");
  }
  | (enumeration_literal) => e1=enumeration_literal
  {
    $expr = new FVhPExprEnumLiteral( $e1.tok.getLine(), $e1.tok.getText(), $e1.isCharLit);
  }
  | numeric_literal
  {
    $expr = $numeric_literal.expr;
  }
  ;


logical_name returns [Token tok;]
  :
  identifier { $tok = $identifier.tok;}
  ;

logical_name_list returns [LinkedList<String> libs, Integer line]
  :
  nm1=logical_name
  {
      $line = $nm1.tok.getLine();
      if ( null == $libs )
      {
          $libs = new LinkedList<String>();
      }
      $libs.add( $nm1.tok.getText() );
  }
  ( COMMA 
        nm2=logical_name
        {
            if ( null == $libs )
              {
                  $libs = new LinkedList<String>();
              }
              $libs.add( $nm2.tok.getText() );
        }
  )*
  ;

logical_operator returns [FVhPExprBinary.OpType op]
  :
  AND { $op = FVhPExprBinary.OpType.AND; }
  | OR { $op = FVhPExprBinary.OpType.OR; }
  | NAND { $op = FVhPExprBinary.OpType.NAND; }
  | NOR { $op = FVhPExprBinary.OpType.NOR; }
  | XOR { $op = FVhPExprBinary.OpType.XOR; }
  | XNOR { $op = FVhPExprBinary.OpType.XNOR; }
  ;

loop_statement
    @init
    {
        String label = null;
        Integer line = -1;
        FVhPSeqLoopStatment.LoopType loopType = FVhPSeqLoopStatment.LoopType.SIMPLELOOP;
        FVhPBase loop = null;
        FVhPLoopParameter param = null;
        FVhPExprBase cond = null;
    }
  :
  (label_colon {label = $label_colon.tok.getText(); })? 
  (
      itr=iteration_scheme
      {
          loopType = $itr.type;
          param = $itr.param;
          cond =  $itr.cond;
      } 
  )? l1=LOOP
  {
  	line = $l1.getLine();
	loop = new FVhPSeqLoopStatment(line, loopType, param, cond );
    ((FVhPSeqLoopStatment)loop).setLabel( label );
    FVhPDescription.instance().addStatement( loop );
    FVhPDescription.instance().pushScope( loop );
  }
  sequence_of_statements 
  END LOOP (identifier)? SEMI
  {
      FVhPDescription.instance().popScope();
  }
  ;


mode returns [FVhPPort.Direction dir]
  :
  IN { $dir = FVhPPort.Direction.INPUT; }
  | OUT { $dir = FVhPPort.Direction.OUTPUT; }
  | INOUT { $dir = FVhPPort.Direction.INOUT; }
  | BUFFER { $dir = FVhPPort.Direction.BUFFER; }
  | LINKAGE { $dir = FVhPPort.Direction.LINKAGE; }
  ;

multiplying_operator returns [FVhPExprBinary.OpType op]
  :
  MUL {$op = FVhPExprBinary.OpType.MUL;}
  | DIV {$op = FVhPExprBinary.OpType.DIV;}
  | MOD  {$op = FVhPExprBinary.OpType.MOD;}
  | REM  {$op = FVhPExprBinary.OpType.REM;}
  ;


// was
//   name
//     : simple_name
//     | operator_symbol
//     | selected_name
//     | indexed_name
//     | slice_name
//     | attribute_name
//     ;
// changed to avoid left-recursion to name (from selected_name, indexed_name,
// slice_name, and attribute_name, respectively)
// (2.2.2004, e.f.)



name returns [Token tok, 
              String fullname, 
              LinkedList<FVhPBase> rangelist, 
              String attribName, 
              boolean isBitString,
              FVhPBase bitsel,
              FVhPBase selectedName,
              FVhPBase aggregate,
              FVhPBase obj,
              FVhPBase declaredObj
              ]
    @init
    {
        FVhPBase objdecl = null;
    }
    :
  (
    nm1=identifier 
    {
        $tok = $nm1.tok; 
        $fullname = $nm1.tok.getText();
        objdecl = FVhPDescription.instance().getSymbol( $nm1.tok.getText() );
        $obj = new FVhPSimpleName($nm1.tok.getLine(), $nm1.tok.getText(), null);
        $declaredObj = objdecl; // Intrroduced to handle part select
    }    
    | 
    STRING_LITERAL 
    {
        $tok = $STRING_LITERAL; 
        $fullname = $tok.getText();
        $isBitString = true;
        objdecl = FVhPDescription.instance().getSymbol( $tok.getText() );
        $obj = new FVhPSimpleName($tok.getLine(), $tok.getText(), null);
        $declaredObj = objdecl;
    }
  )
  ( options {greedy=true;}:
    (
      DOT nm2=suffix 
      {
        FVhPBase ele = null;
        if ( null == $selectedName )
        {
            if ( null != objdecl )
            {
        	   $selectedName = new FVhPSelectedName(  $DOT.getLine(), objdecl );
        	   ((FVhPSelectedName)$selectedName).addElement( objdecl );
        	}
        	else
        	{
        	    $selectedName = new FVhPSelectedName(  $DOT.getLine(), $obj );
                ((FVhPSelectedName)$selectedName).addElement( $obj );
        	}
        }	 
        if ( null != $nm2.tok )
        {
         	 $fullname += "." + $nm2.tok.getText();
         	 if ( null != objdecl && FVhPBase.ObjType.TYPEDECL == objdecl.getObjType() )
         	 {
         	 	ele = ((FVhPRecordType)objdecl).getElement( $nm2.tok.getText() );
         	 	objdecl = ele;
         	 	$declaredObj = objdecl;
         	 }
         	 else
         	 {      
             	ele = new FVhPSimpleName( $nm2.tok.getLine(), $nm2.tok.getText(), null);
             }
        }
        else if ( null != $nm2.all )
        {
             $fullname += "." + $nm2.all ;
             ele = new FVhPSimpleName( $nm1.tok.getLine(), $nm2.all, null);
        }
        ((FVhPSelectedName)$selectedName).addElement( ele );
        $obj = $selectedName;
      }
      | 
      APOSTROPHE aggregate
      {
          $obj = null;
          $aggregate = $aggregate.aggr;
      }
      | 
      (
          sig=signature
          {
              $obj = null;
              //FVhPSignature signature
              LogMessages.instance().print(3576, LogMessages.MSG_SEVERITY.ERROR, "attribute with signature is not yet supported ...");
          }
      )?
      ap=APOSTROPHE attribute_designator
      {
          $attribName = $attribute_designator.attribName ;
//          $obj = new FVhPSimpleName($ap.getLine(), $attribName, null);
      }
      | (LPAREN expression (COMMA expression)* RPAREN) => l1=LPAREN e1=expression 
      {
          if ( null != $bitsel ) // Multi-dimensional array e.g sig1 <= ARR(0)(2);
          {
              objdecl = $bitsel;
              $declaredObj = objdecl;
          }
          if ( null != objdecl )
          {
              $bitsel = new FVhPBitSelect($l1.getLine(), objdecl, null);
              ((FVhPBitSelect)$bitsel).addIndex( $e1.expr );
              $declaredObj = $bitsel;
          }
          else
          {
          	$bitsel = new FVhPBitSelect($l1.getLine(), $obj, null);
            ((FVhPBitSelect)$bitsel).addIndex( $e1.expr ); 
            $declaredObj = $bitsel;         
          }
      }
      (
          COMMA e2=expression
          {
              ((FVhPBitSelect)$bitsel).addIndex( $e2.expr );
          }
      )* 
      RPAREN
//      {
//          $obj = $bitsel;
//      }
      | (LPAREN actual_parameter_part RPAREN) => l1=LPAREN actual_parameter_part RPAREN
      {
          LogMessages.instance().print(3575, LogMessages.MSG_SEVERITY.ERROR,"actual_parameter_part are not yet supported ...  File " + 
                      FVhPDescription.instance().getFile() + ":" + $l1.getLine());
          $obj = null;            
      }
      | 
      lp1=LPAREN dr1=discrete_range
      {
          $rangelist  = new LinkedList<FVhPBase>();
          $rangelist.add($dr1.rng);
      }
      (COMMA dr2=discrete_range
      {
          $rangelist.add($dr2.rng);
        }
      )* 
      RPAREN
//      {
//          if ( null != objdecl && FVhPBase.ObjType.SUBTYPEIND == objdecl.getObjType() )
//          {
//              $obj = new FVhPSubTypeIndication($lp1.getLine(), objdecl, $rangelist );
//          }
//      }
    )
  )*
  ;


next_statement
  :
  (label_colon)? NEXT (identifier)? (WHEN condition)? SEMI
  ;

numeric_literal returns [FVhPExprBase expr]
  :
  (al1=abstract_literal) => al2=abstract_literal
  {
      $expr = $al2.expr;
  }
  | physical_literal
  {
      $expr = $physical_literal.expr; 
  }
  ;

object_declaration
  :
  constant_declaration
  | signal_declaration
  | variable_declaration
  | file_declaration
  ;

opts
  :
  (GUARDED)? (delay_mechanism)?
  ;

//package_body
//  : PACKAGE BODY identifier IS
//    package_body_declarative_part
//    END ( PACKAGE BODY )? ( identifier )? SEMI
//  ;

package_body returns[  FVhPPackageBody packbody ]
    @init
    {
        String name = null;
    }
  :
  PACKAGE BODY id=identifier IS
  {
  	FVhPDescription desc = FVhPDescription.instance();
    name = $id.tok.getText();  
    try
    {
      	FVhPPackage pack = (FVhPPackage)desc.restoreDesignUnit( workLib, name );
    }
    catch (IOException e) 
    {
    	e.printStackTrace();
    }
    
    FVhPBase packdecl = desc.getSymbol( name );
    $packbody = new FVhPPackageBody( name,  $id.tok.getLine() );
    ((FVhPPackageBody)$packbody).setPackage( (FVhPPackage)packdecl );
    desc.insertSymbol( name, $packbody);
    desc.pushScope($packbody);      
  } 
  package_body_declarative_part 
  END (PACKAGE BODY)? (nm2=identifier )? SEMI
  {
    FVhPDescription desc = FVhPDescription.instance();
    desc.popScope();
    FVhPDescription.instance().getRoot().addDesignUnit($packbody);
    libs = null;
    uses = null;
  }
  ;

package_body_declarative_item
  :
  (subprogram_declaration) => subprogram_declaration
  | subprogram_body
  | type_declaration
  | subtype_declaration
  | constant_declaration
  | variable_declaration
  | file_declaration
  | alias_declaration
  | use_clause
  | group_template_declaration
  | group_declaration
  ;

package_body_declarative_part
  :
  (package_body_declarative_item)*
  ;

//package_declaration
//  : PACKAGE identifier IS
//    package_declarative_part
//    END ( PACKAGE )? ( identifier )? SEMI
//  ;

package_declaration returns [FVhPPackage pack]
  @init 
  { 
      String name = null;
  }
  :
  PACKAGE id=identifier IS
  {
      name = $id.tok.getText();
    FVhPDescription desc = FVhPDescription.instance();
    $pack = new FVhPPackage( name,  $id.tok.getLine() );
    desc.insertSymbol( name, $pack);
    desc.pushScope($pack);
  }
  package_declarative_part 
  END (PACKAGE)? (nm2=identifier)? SEMI
  {
    FVhPDescription desc = FVhPDescription.instance();
     desc.popScope();
     FVhPDescription.instance().getRoot().addDesignUnit($pack);
//      $pack.setLibraries(libs);
//      $pack.setUseClauses(uses);
      libs = null;
      uses = null;
  }
  ;

package_declarative_item
  :
  subprogram_declaration
  | type_declaration
  | subtype_declaration
  | constant_declaration
  | signal_declaration
  | variable_declaration
  | file_declaration
  | alias_declaration
  | component_declaration
  | attribute_declaration
  | attribute_specification
  | disconnection_specification
  | use_clause
  | group_template_declaration
  | group_declaration
  ;

package_declarative_part
  :
  (package_declarative_item)*
  ;

parameter_specification returns [ FVhPLoopParameter param, FVhPForIndex forindex ]
  :
  identifier IN discrete_range
  {
      Integer line = $identifier.tok.getLine();
      FVhPIdentifier  id = new FVhPIdentifier( line, $identifier.tok.getText() );
      FVhPForIndex index = new FVhPForIndex( line, id,   $discrete_range.rng);
      $forindex = index;
      FVhPDescription.instance().addSymbol( $identifier.tok.getText(), index );
      $param = new FVhPLoopParameter( $identifier.tok.getLine(), 
                                      id, 
                                      $discrete_range.rng );
      index.setLoopParameter( $param );
      $param.setLoopIndex( index );                      
  }
  ;

physical_literal returns [FVhPExprBase expr]
    @init
    {
        FVhPExprBase val = null;
    }
  :
  (abstract_literal { val = $abstract_literal.expr; })? name
  {
      FVhPBase unitdef = FVhPDescription.instance().getSymbol( $name.tok.getText() ); 
      if ( null != unitdef )
      {
      	$expr = new FVhPExprPhysicalLiteral($name.tok.getLine(), val, unitdef );
      }
      else if ( true == $name.isBitString )
      {
      	String str = $name.tok.getText();
        String valWithoutQuotes = str.substring(1, str.length()-1);
        if ( valWithoutQuotes.matches("[^01xXzZ_]+") )
        {
          $expr = new FVhPExprBitString($name.tok.getLine(), valWithoutQuotes); 
        }
        else
        {
          $expr = new FVhPExprBitString($name.tok.getLine(), valWithoutQuotes); 
        }    
      }
      else
      {
      	LogMessages.instance().print(3574, LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR, "Unknown literal found as expression ..." );
      }
  }  
  ;

physical_type_definition returns [FVhPTypeDefBase typedef]
  :
  range_constraint u1=UNITS base_unit_declaration
  {
      FVhPBase range = $range_constraint.rng;
      FVhPBase base = $base_unit_declaration.baseunit;
      $typedef = new FVhPPhysicalType( range.getLine(), range, (FVhPPhysicalTypeBaseUnit)base );
  }
   (
       secondary_unit_declaration
       {
           ((FVhPPhysicalType)$typedef).addSecondaryUnit( (FVhPPhysicalTypeSecondaryUnit)$secondary_unit_declaration.secunit );
       }
   )* END UNITS (identifier)?
  ;

port_clause returns [LinkedList<FVhPPort> ports]
  :
  PORT LPAREN port_list RPAREN SEMI
  {
          $ports = new LinkedList<FVhPPort>();
          LinkedList<FVhPInterfaceElement> intfs = $port_list.intf_ele;
            if ( intfs != null && false == intfs.isEmpty() )
        {
            Iterator itr = intfs.iterator(); 
            while(itr.hasNext()) 
            {
                FVhPInterfaceElement intf = (FVhPInterfaceElement)itr.next(); 
                FVhPPort port = new FVhPPort(intf);
                $ports.add(port);
                String portName = intf.getIdentifier().getName();
                FVhPDescription desc = FVhPDescription.instance(); 
                desc.insertSymbol( portName, port );
            } 
        }
  }
  ;

port_list returns [LinkedList<FVhPInterfaceElement> intf_ele]
  :
  interface_list
  {
      $intf_ele = $interface_list.intf_ele;
  }
  ;

port_map_aspect returns[LinkedList<FVhPPortMap> portmaps]
  :
  PORT MAP LPAREN association_list RPAREN
  {
  	$portmaps = $association_list.portmaps;
   	if ( null == $portmaps && null != $association_list.actuals && 
  	     false == $association_list.actuals.isEmpty() )
  	{
  		LinkedList<FVhPPort> ports = getCurrentInstanceMasterPortList();
  		Iterator<FVhPPort> itr2 = ( null != ports && false == ports.isEmpty() ) ? ports.iterator() : null; 
  		$portmaps = new LinkedList<FVhPPortMap>();
		Iterator<FVhPExprBase> itr = $association_list.actuals.iterator();
		while ( itr.hasNext() )
		{
			FVhPExprBase expr = itr.next();
			FVhPPort  formal = ( null != itr2 ) ? itr2.next() : null ;
			FVhPPortMap map = new FVhPPortMap( $LPAREN.getLine(), formal, expr );
			$portmaps.add( map );
		}
  	} 	
  }
  ;

primary returns [FVhPExprBase expr]
  :
  (function_call) => f1=function_call
  {
      $expr = $f1.expr;
  }
  | (name APOSTROPHE) => qualified_expression
  {
//    LogMessages.instance().print(3573, LogMessages.MSG_SEVERITY.ERROR, "Qualified expression is not supported yet ..." );
    $expr = $qualified_expression.expr;
  }
  | (LPAREN expression RPAREN) => LPAREN expression RPAREN
  {
      $expr = $expression.expr;
  }
  | literal
  {
    $expr = $literal.expr;
  }
  | allocator
  {
//    LogMessages.instance().print(3572, LogMessages.MSG_SEVERITY.ERROR, "Allocator as expression is not supported yet ..." );
    $expr = $allocator.expr;
  }
  | aggregate
  {
      $expr = new FVhPExprAggregate( $aggregate.aggr.getLine(), $aggregate.aggr);
  }
  ;

primary_unit returns [FVhPBase du]
  :
  entity_declaration
  {
      $du = $entity_declaration.entity;
  }
  | configuration_declaration
  {
      $du = $configuration_declaration.du;
  }
  | package_declaration
  {
      $du = $package_declaration.pack;
  }
  ;
  
procedural_declarative_item
  :
  (subprogram_declaration) => subprogram_declaration
  | subprogram_body
  | type_declaration
  | subtype_declaration
  | constant_declaration
  | variable_declaration
  | alias_declaration
  | attribute_declaration
  | attribute_specification
  | use_clause
  | group_template_declaration
  | group_declaration
  ;

procedural_declarative_part
  :
  (procedural_declarative_item)*
  ;

procedural_statement_part
  :
  (sequential_statement)*
  ;

procedure_call returns[FVhPBase proccall]
	@init
	{
		FVhPBase subp = null;
		FVhPDescription desc = FVhPDescription.instance();
	}
  :
  name
  {
  	subp =  desc.getSymbol( $name.tok.getText() );
  	currSubprog = subp;
  } 
  (LPAREN actual_parameter_part RPAREN)?
  {
  	
  	FVhPBase.ObjType objtype = ( null != subp ) ? subp.getObjType()  : FVhPBase.ObjType.UNKNOWN ;
  	if ( ( null != $name.bitsel || null == subp ) || 
  	     ( FVhPBase.ObjType.SUBPROGRAM_DECL == objtype ) ||
  	     ( FVhPBase.ObjType.SUBPROGRAM_BODY == objtype )  )
  	{
  	  	$proccall = new FVhPSubProgramCall( $name.tok.getLine(), $name.tok.getText() );
  	  	((FVhPSubProgramCall)$proccall).setSubprogramDefinition( subp );
  	  	if ( null != $actual_parameter_part.paramlist )
  	  	{
  	  		((FVhPSubProgramCall)$proccall).setArguments($actual_parameter_part.paramlist);
  	  	}
  	  	else if ( null != $name.bitsel )
  	  	{
  	  		 LinkedList<FVhPExprBase> args = ((FVhPBitSelect) $name.bitsel).getIndexList();
  	  		 ((FVhPSubProgramCall)$proccall).setArguments( args );
  	  	}
  	}
	else if ( FVhPBase.ObjType.COMPONENTDECL == objtype )
	{
		$proccall = new FVhPInstantiation($name.tok.getLine(), $name.tok.getText(), subp );
	}
	else
	{
		LogMessages.instance().print(3571, LogMessages.MSG_SEVERITY.PROGRAMMING_ERROR, "Unsupprted object type got considered as procedure call for line " + $name.tok.getLine());
	}
	currSubprog = null;  
  }
  
  ;

procedure_call_statement returns[FVhPBase stmt]
	@init
	{
		String label = null;
	}
  :
  (label_colon 
  {
  	label = $label_colon.tok.getText();
  } 
  )? procedure_call SEMI
  {
  	$stmt = $procedure_call.proccall;
  	((FVhPSubProgramCall)$stmt).setLabel( label );
  	FVhPDescription.instance().addStatement( $stmt );
  }  
  ;

process_declarative_item
  :
  (subprogram_declaration) => subprogram_declaration
  | subprogram_body
  | type_declaration
  | subtype_declaration
  | constant_declaration
  | variable_declaration
  | file_declaration
  | alias_declaration
  | attribute_declaration
  | attribute_specification
  | use_clause
  | group_template_declaration
  | group_declaration
  ;

process_declarative_part
  :
  (process_declarative_item)*
  ;

process_statement returns [FVhPBase process]
  @init
  {
      Integer line = -1;
      String label = null;
      FVhPDescription desc = FVhPDescription.instance();
  }
  @after
  {
      desc.popScope();
      desc.addStatement( $process );
  }
  :
  (label_colon 
  { 
      label = $label_colon.tok.getText(); 
      line = $label_colon.tok.getLine();
  })? 
  (POSTPONED)? p1=PROCESS
  {
      line = $p1.getLine();
      $process = new FVhPProcessStmt(line);
    ((FVhPProcessStmt)$process).setLabel( label );
    desc.pushScope($process);
  } 
  (
      LPAREN s1=sensitivity_list RPAREN
      {
          ((FVhPProcessStmt)$process).setSensitivityList( $s1.sensitivity_list );
      }
  )? (IS)? 
  process_declarative_part
  BEGIN 
  process_statement_part
  END (POSTPONED)? PROCESS (identifier)? SEMI
  ;

process_statement_part
  :
  (sequential_statement)*
  ;

qualified_expression returns [FVhPExprBase expr]
    @init
    {
        FVhPBase typemark = null;
    }
  :
  nm1=name
  {
      String typeName = getNameString( $nm1.obj );
      if ( null == typeName )
      {
        typeName = $nm1.tok.getText();
      }
      if ( null != typeName )
      {
        typemark = FVhPDescription.instance().getSymbol( typeName );
      }
      if ( null == typemark && $nm1.obj instanceof FVhPSelectedName )
      {
          typemark = $nm1.obj;
      }
      if ( null == typemark )
      {
        LogMessages.instance().print(3559, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+
                                      $nm1.tok.getText() +
                                      "' in " + FVhPDescription.instance().getFile() +
                                      ":" + $nm1.tok.getLine() );
      }
   }
   APOSTROPHE
  (
    (aggregate) => aggr=aggregate
    {
        $expr = new FVhPExprQualified( $nm1.tok.getLine(), typemark, $aggr.aggr );
    }
    | 
    LPAREN expression RPAREN
    {
    $expr = new FVhPExprQualified( $nm1.tok.getLine(), typemark, $aggr.aggr );
    }
  )
  {
      LogMessages.instance().print(3570, LogMessages.MSG_SEVERITY.WARNING, "Ignoring qualified exprression found at " +  
                 FVhPDescription.instance().getFile() + 
                 ":" +  $APOSTROPHE.getLine() + " ...");
  } 
  ;


range returns [FVhPBase rng]
    @init
    {
        FVhPExprBase lt = null;
          FVhPExprBase rt = null;
    }
  :
  (simple_expression direction simple_expression) => (e1=simple_expression direction e2=simple_expression)
  {
      lt = $e1.expr;
      rt = $e2.expr;
      $rng = new FVhPRange(lt, rt, $direction.dir, ((lt != null) ? lt.getLine() : 0 ) );
  }
  | name
  {
    FVhPBase typemark = FVhPDescription.instance().getSymbol( $name.tok.getText() );
    if ( null == typemark )
    {
        LogMessages.instance().print(3569, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+
                                      $name.tok.getText() + 
                                      "' in " + FVhPDescription.instance().getFile() +
                                      ":" + $name.tok.getLine() );
    }
    if ( null != $name.attribName )
    {
        $rng = new FVhPExprAttribute($name.tok.getLine(), typemark, $name.attribName);
    }
    else
    {
        $rng = new FVhPSubTypeIndication($name.tok.getLine(), typemark, $name.rangelist);
    }
  }
  ;

range_constraint returns [FVhPBase rng]
  :
  RANGETOK range
  {
      $rng = $range.rng; 
  }
  ;

record_nature_definition
  :
  RECORD + END RECORD (identifier)?
  ;

record_type_definition returns [FVhPRecordType def]
    @init
    {
        LinkedList<FVhPRecordElement> recelelist = null;
        $def = new FVhPRecordType( recelelist, -1);
        FVhPDescription desc = FVhPDescription.instance();
        desc.pushScope($def);
        Integer line = null; 
    }
    @after
    {
        desc.popScope();
        $def.setElements( recelelist );
        $def.setLine( line );
    }
  :
  RECORD 
  (
      ele=element_declaration
      {
          if ( null == recelelist )
          {
              recelelist = new LinkedList<FVhPRecordElement>();
          }
          LinkedList<Token> ids = $ele.ids;    
        if ( null != ids &&  false == ids.isEmpty() )
         {
            Iterator itr = ids.iterator(); 
            while(itr.hasNext()) 
            {
                Token tok = (Token)itr.next(); 
                String elename = tok.getText();
                if ( null == line )
                {
                    line = tok.getLine();
                }
                FVhPIdentifier ident = new FVhPIdentifier( tok.getLine(), elename );
                FVhPRecordElement element = new FVhPRecordElement ( tok.getLine(), ident, $ele.sind );
                recelelist.add(element);
            }
        }      
      }
  )+ 
  END RECORD (identifier)?
  ;


relation returns [FVhPExprBase expr]
    @init
    {
        FVhPExprBase lhs = null;
        FVhPExprBase rhs = null;
        FVhPExprBinary.OpType op = FVhPExprBinary.OpType.UNKNOWN;
    }
  :
  e1=shift_expression
  {
          lhs = $e1.expr;
  } 
          (options {greedy=true;}: 
              relational_operator
              {
                  op = $relational_operator.op;
              } 
              e2=shift_expression
              {
                  rhs = $e2.expr;
              }
          )?
  {
      if ( null != rhs )
      {
          if ( null == lhs )
          {
              LogMessages.instance().print(3568, LogMessages.MSG_SEVERITY.ERROR, "In the rule shift_expression the lhs is null ...");
          }
          else
          {
              $expr = new FVhPExprBinary ( ( null == lhs ) ? 0 : lhs.getLine(), lhs, rhs, op );
          } 
      }
      else
      {
          $expr = lhs;
      }
  }
  ;

relational_operator returns [FVhPExprBinary.OpType op]
  :
  EQ { $op = FVhPExprBinary.OpType.EQ; }
  | NEQ { $op = FVhPExprBinary.OpType.NEQ; }
  | LOWERTHAN { $op = FVhPExprBinary.OpType.LOWERTHAN; }
  | LE { $op = FVhPExprBinary.OpType.LE; }
  | GREATERTHAN { $op = FVhPExprBinary.OpType.GREATERTHAN; }
  | GE { $op = FVhPExprBinary.OpType.GE; }
  ;

report_statement
  :
  (label_colon)? REPORT expression (SEVERITY expression)? SEMI
  {
      LogMessages.instance().print(3567, LogMessages.MSG_SEVERITY.WARNING, "Ignoring report statement found at " +  
                 FVhPDescription.instance().getFile() + 
                 ":" +  $REPORT.getLine() + " ...");
  }
  ;

return_statement returns [FVhPBase retstmt]
    @init
    {
        String label = null;
        FVhPExprBase retexpr = null;
    }
  :
  (
      label_colon
      {
          label = $label_colon.tok.getText();
      }
  )? 
  RETURN 
  (
      expression
      {
          retexpr = $expression.expr;
      }
  )? 
  SEMI
  {
      $retstmt = new FVhPReturnStatement( $RETURN.getLine(), retexpr );
      ((FVhPReturnStatement)$retstmt).setLabel( label );
      FVhPDescription.instance().addStatement( $retstmt );
  }
  ;


scalar_type_definition returns [FVhPTypeDefBase type  ]
  :
  (range_constraint UNITS) => phy=physical_type_definition
  {
      $type = $phy.typedef;
  }
  | 
  en=enumeration_type_definition
  {
      $type = new FVhPEnumType($en.literals, $en.line, $en.isCharLit );
      Iterator<FVhPEnumLiteral> itr = $en.literals.iterator();
      while ( itr.hasNext() )
      {
          FVhPEnumLiteral lit = itr.next();
          lit.setType( (FVhPEnumType)$type );
      }
  }
  | 
  r1=range_constraint
  {
      $type = new FVhPRangeContraintType( $r1.rng, $r1.rng.getLine() );
  }
  ;

secondary_unit returns [FVhPBase du]
  :
  architecture_body
  {
      $du = $architecture_body.arch;
  }
  | package_body
  {
      $du = $package_body.packbody;
  } 
  ;

secondary_unit_declaration returns [FVhPBase secunit]
  :
  identifier EQ physical_literal SEMI
  {
      $secunit = new FVhPPhysicalTypeSecondaryUnit( $identifier.tok.getLine(), 
                                                  $identifier.tok.getText(),
                                                  $physical_literal.expr );
    FVhPDescription.instance().addSymbol( $identifier.tok.getText(), $secunit );                                                  
  }
  ;

selected_signal_assignment returns [FVhPBase assgn]
  :
  WITH expression SELECT target LE opts selected_waveforms SEMI
  {
  	$assgn = new FVhPSelectedSignalAssignment( $WITH.getLine(),$expression.expr, $target.target, 
			                                   $selected_waveforms.selectedWvfrms);
	FVhPDescription.instance().addStatement( $assgn );	                               
  } 
  ;

selected_waveforms returns[LinkedList<FVhPWaveFormWithChoices> selectedWvfrms]
  :
  w1=waveform WHEN c1=choices 
  {
  	$selectedWvfrms = new LinkedList<FVhPWaveFormWithChoices> ();
  	FVhPWaveFormWithChoices wvfrm = new FVhPWaveFormWithChoices( $w1.wvfrm.getLine(), $w1.wvfrm, $c1.choices );
  	$selectedWvfrms.add( wvfrm );
  }
  (
  	COMMA w2=waveform WHEN c2=choices
  	{
		FVhPWaveFormWithChoices wvfrm = new FVhPWaveFormWithChoices( $w2.wvfrm.getLine(), $w2.wvfrm, $c2.choices );
  		$selectedWvfrms.add( wvfrm );
  	}
  )*
  ;

sensitivity_clause returns [LinkedList<FVhPBase> sensitivity_list]
  :
  ON sensitivity_list
  {
      $sensitivity_list = $sensitivity_list.sensitivity_list;
  }
  ;

sensitivity_list returns [LinkedList<FVhPBase> sensitivity_list]
    @init
    {
        FVhPDescription desc = FVhPDescription.instance();
    }
  :
  n1=name 
  {
    $sensitivity_list = new LinkedList<FVhPBase>();
    String signame = getNameString ( $n1.obj );
    FVhPBase sym = (FVhPBase)desc.getSymbol( signame );
    if ( null == sym && $n1.obj instanceof FVhPSelectedName )
    {
        sym = $n1.obj;
    }
    if ( null == sym  )
    {
        LogMessages.instance().print(3566, LogMessages.MSG_SEVERITY.ERROR, "Found reference to undefined symbol '"+ 
                                       signame + 
                                       "' in the process sensitivity list at " +
                                       desc.getFile() +
                                       ":" + $n1.tok.getLine() );
    }
    else
    {
        switch ( sym.getObjType() )
        {
            case SIGNAL: 
                $sensitivity_list.add (sym );
                break;
            case PORT:
                $sensitivity_list.add (sym );
                break;
            case INTERFACE_ELEMENT:
                $sensitivity_list.add (sym );
                break;
            case SELECTEDNAME:
                $sensitivity_list.add (sym );
                break;
            default:
                LogMessages.instance().print(3565, LogMessages.MSG_SEVERITY.ERROR, "The object '"+ signame + 
                                       "' in the process sensitivy list at " +
                                       desc.getFile() +
                                       ":" + $n1.tok.getLine() + " is not a signal ..." );
                break;
        }
    }
  } 
  (
  COMMA n2=name
  {
    String signame = getNameString ( $n2.obj );
    FVhPBase sym = (FVhPBase)desc.getSymbol( signame );
    if ( null == sym && $n2.obj instanceof FVhPSelectedName )
    {
        sym = $n2.obj;
    }
    if ( null == sym  )
    {
        LogMessages.instance().print(3564, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+ 
                                      signame + 
                                      "' in the process sensitivy list at " +
                                      desc.getFile() +
                                      ":" + $n2.tok.getLine() );
    }
    else
    {
        switch ( sym.getObjType() )
        {
            case SIGNAL: 
                $sensitivity_list.add (sym );
                break;
            case PORT:
                $sensitivity_list.add (sym );
                break;
            case INTERFACE_ELEMENT:
                $sensitivity_list.add (sym );
                break;
            case SELECTEDNAME:
                $sensitivity_list.add (sym );
                break;    
            default:
                LogMessages.instance().print(3563, LogMessages.MSG_SEVERITY.ERROR, "The object '"+ signame +
                                       "' in the process sensitivy list at " +
                                       desc.getFile() +
                                       ":" + $n2.tok.getLine() + " is not a signal ..." );
                break;
        }
    }
  }
  )*
  ;

sequence_of_statements
  :
  (sequential_statement)*
  ;

sequential_statement
    @init
    {
        String label = null;
    }
  :
  ( (label_colon)? WAIT) => wait_statement
  | ( (label_colon)? ASSERT) => assertion_statement
  | ( (label_colon)? REPORT) => report_statement
  | ( (label_colon)? target LE) => signal_assignment_statement
  | ( (label_colon)? target VARASGN) => variable_assignment_statement
  | ( (label_colon)? IF) => if_statement
  | ( (label_colon)? CASE) => case_statement
  |
  (
    (label_colon)? (iteration_scheme)? LOOP
  )
    => loop_statement
  | ( (label_colon)? NEXT) => next_statement
  | ( (label_colon)? EXIT) => exit_statement
  | ( (label_colon)? RETURN) => return_statement
  | ( (label_colon)? NULLTOK SEMI) => (label_colon{ label = $label_colon.tok.getText();})? n1=NULLTOK SEMI
  {
      FVhPNullStatement nullstmt = new FVhPNullStatement($n1.getLine());
      nullstmt.setLabel ( label );
      FVhPDescription.instance().addStatement ( nullstmt );
  }
  | procedure_call_statement
  ;

shift_expression returns [FVhPExprBase expr]
    @init
    {
        FVhPExprBase lhs = null;
        FVhPExprBase rhs = null;
        FVhPExprBinary.OpType op = FVhPExprBinary.OpType.UNKNOWN;
    }
  :
  e1=simple_expression
  {
      lhs = $e1.expr;
  } 
          (options {greedy=true;}: 
              shift_operator
              {
                  op = $shift_operator.op;
              } 
              e2=simple_expression
              {
                  rhs = $e2.expr;
              }
          )?
  {
      if ( null != rhs )
      {
          $expr = new FVhPExprBinary( ( null == lhs ) ? 0 : lhs.getLine(), lhs, rhs, op );
      }
      else
      {
          $expr = lhs;
      }
  }
  ;

shift_operator returns [FVhPExprBinary.OpType op]
  :
  SLL { $op = FVhPExprBinary.OpType.SLL; }
  | SRL { $op = FVhPExprBinary.OpType.SRL; }
  | SLA { $op = FVhPExprBinary.OpType.SLA; }
  | SRA { $op = FVhPExprBinary.OpType.SRA; }
  | ROL { $op = FVhPExprBinary.OpType.ROL; }
  | ROR { $op = FVhPExprBinary.OpType.ROR; }
  ;


signal_assignment_statement
	@init
    {
    	String label = null;
    }
  :
  (label_colon{label = $label_colon.tok.getText(); })? 
  target
  {
    setCurrentTarget(  $target.target );
  } 
  LE 
  (delay_mechanism)? waveform SEMI
  {
        setCurrentTarget(  null );
        LinkedList<FVhPWaveForm> wvforms = new LinkedList<FVhPWaveForm>();
        wvforms.add( $waveform.wvfrm );
        FVhPSignalAssignment assmt = new FVhPSignalAssignment($target.target.getLine(), $target.target, 
                                                  wvforms);
        assmt.setLabel(label);                                          
        FVhPDescription.instance().addStatement(assmt);
  }
  ;

signal_declaration returns [LinkedList<FVhPSignal> sigs]
    @init
    {
        LinkedList<Token> ids = null;
        FVhPSubTypeIndication sind = null;
        FVhPExprBase initval = null;
    }
    @after
    {
        myCurrentSind = null;
    }
  :
  SIGNAL idents=identifier_list{ ids = $idents.ids;} COLON si=subtype_indication{sind = $si.sind; myCurrentSind = $si.sind; }
  (signal_kind)? (VARASGN expression { initval = $expression.expr;})? SEMI
  {
       $sigs  =  new LinkedList<FVhPSignal>();
       if ( null != ids &&  false == ids.isEmpty() )
     {
        Iterator itr = ids.iterator(); 
        while(itr.hasNext()) 
        {
            Token tok = (Token)itr.next(); 
            String signame= tok.getText();
            
            FVhPIdentifier ident = new FVhPIdentifier( tok.getLine(), signame );
            FVhPSignal sig = new FVhPSignal ( tok.getLine(), ident, sind );
            FVhPDescription desc = FVhPDescription.instance();
            FVhPBase sym = (FVhPBase)desc.getSymbol( signame );
            if ( null != sym  && desc.getScope() == sym.getScope() )
            {
                LogMessages.instance().print(3562, LogMessages.MSG_SEVERITY.ERROR, "Found already defined symbol '" + 
                                    signame + "' reference in " +
                                       desc.getFile() +
                                       ":" + tok.getLine() +
                                       " in file " + sym.getFile() + ":"  + sym.getLine() );
            } 
            desc.insertSymbol( signame, sig );
            desc.addDeclaration(sig);
            if ( null != initval )
            {
                initval.setSubtypeIndication( sind );
            }
            sig.setInitialValue(initval);
            $sigs.add(sig);
        }
    }
   } 
  ;

signal_kind
  :
  REGISTER
  | BUS
  ;

signal_list
  :
  name  (COMMA name)*
  | OTHERS
  | ALL
  ;

signature returns [FVhPSignature signature]
    @init
    {
        String objName = null;
    }
  :
  LBRACKET 
  {
      $signature = new FVhPSignature( $LBRACKET.line );
  }
  (
      nm1=name
      {
        objName = getNameString ( $nm1.obj );
          $signature.addArgument( objName );
      } 
      (
    COMMA nm2=name
    {
        objName = getNameString ( $nm2.obj );
        $signature.addArgument( objName );
    }
      )*
  )? 
  (
  RETURN nm3=name
  {
    objName = getNameString ( $nm3.obj );
      $signature.setReturnTypeName( objName );
  }
  )? 
  RBRACKET
  ;

// NOTE that sign is applied to first operand only (LRM does not permit
// `a op -b' - use `a op (-b)' instead).
// (3.2.2004, e.f.)


simple_expression returns [FVhPExprBase expr]
  @init
  {
      boolean isPlus = false;
      boolean isMinus = false;
      FVhPExprBase lhs = null;
      FVhPExprBase rhs = null;
      FVhPExprBinary.OpType op = FVhPExprBinary.OpType.UNKNOWN;
  }
  :
  (
    PLUS { isPlus = true;}
    | MINUS { isMinus = true; }
  )?
  t1=term
  {
      lhs = $t1.expr;
  } 
  (options {greedy=true;}: 
      adding_operator
      {
          op = $adding_operator.op;
      } 
      t2=term
      {
          rhs = $t2.expr;
      }
  )*
  {
      if ( null != rhs )
      {
          $expr = new FVhPExprBinary( ( null == lhs ) ? 0 : lhs.getLine(), lhs, rhs, op );
      }
      else
      {
          $expr = lhs;
      }
      if ( isPlus )
      {
          $expr = new FVhPExprUnary( ( null == lhs ) ? 0 : lhs.getLine(), $expr, FVhPExprUnary.OpType.PLUS);
      }
      if ( isMinus )
      {
          $expr = new FVhPExprUnary( ( null == lhs ) ? 0 : lhs.getLine(), $expr, FVhPExprUnary.OpType.MINUS);
      }
  }
  ;


simple_simultaneous_statement
  :
  (label_colon)? simple_expression ASSIGN simple_expression  SEMI
  ;

simultaneous_alternative
  :
  WHEN choices ARROW simultaneous_statement_part
  ;

simultaneous_case_statement
  :
  (label_colon)? CASE expression USE (simultaneous_alternative)+ END CASE (identifier)? SEMI
  ;

simultaneous_if_statement
  :
  (label_colon)? IF condition USE simultaneous_statement_part (ELSIF condition USE simultaneous_statement_part)* (ELSE simultaneous_statement_part)? END USE (identifier)? SEMI
  ;

simultaneous_statement
    @init
    {
        String label = null;
    }
  :
  ( (label_colon)? simple_expression ASSIGN) => simple_simultaneous_statement
  | ( (label_colon)? IF condition USE) => simultaneous_if_statement
  | ( (label_colon)? CASE expression USE) => simultaneous_case_statement
  | (label_colon{label = $label_colon.tok.getText();})? n1=NULLTOK SEMI
  {
    FVhPNullStatement nullstmt = new FVhPNullStatement($n1.getLine());
      nullstmt.setLabel ( label );
      FVhPDescription.instance().addStatement ( nullstmt );
  }
  ;

simultaneous_statement_part
  :
  (simultaneous_statement)*
  ;

subprogram_body returns [FVhPSubProgramBody subp]
  :
  spec=subprogram_specification IS
  {
      $subp = new FVhPSubProgramBody( $spec.subpspec.getName(), $spec.subpspec.getLine());
      $subp.setInterfaceElementList( $spec.subpspec.getInterfaceElementList() );
      $subp.setReturnType( $spec.subpspec.getReturnType() );
      $subp.setFunction( $spec.subpspec.isFunction() );
      $subp.setSymbolTable( $spec.subpspec.getSymbolTable() );
      FVhPDescription desc = FVhPDescription.instance();
      desc.popScope();
      desc.insertSymbol( $subp.getName(), $subp);
      desc.addDeclaration($subp);
      desc.pushScope($subp);
  } 
  subprogram_declarative_part 
  BEGIN subprogram_statement_part END (subprogram_kind)? (designator)? SEMI
  {
      FVhPDescription.instance().popScope();
  }
  ;

subprogram_declaration
  :
  spec=subprogram_specification SEMI
  {
      FVhPDescription desc = FVhPDescription.instance();
      desc.popScope();
      desc.insertSymbol( $spec.subpspec.getName(), $spec.subpspec);
      desc.addDeclaration($spec.subpspec);
  }
  ;

subprogram_declarative_item
  :
  (subprogram_declaration) => subprogram_declaration
  | subprogram_body
  | type_declaration
  | subtype_declaration
  | constant_declaration
  | variable_declaration
  | file_declaration
  | alias_declaration
  | attribute_declaration
  | attribute_specification
  | use_clause
  | group_template_declaration
  | group_declaration
  ;

subprogram_declarative_part
  :
  (subprogram_declarative_item)*
  ;

subprogram_kind
  :
  PROCEDURE
  | FUNCTION
  ;

subprogram_specification returns [FVhPSubProgramDecl subpspec]
    @init 
    {
        LinkedList<FVhPInterfaceElement> intf_ele = null;
        FVhPSubProgramDecl subp = null;
        isInSubProgramScope = true;
    }
    @after
    {
        if ( null != subp )
        {
            subp.setInterfaceElementList( intf_ele );
        }
        $subpspec = subp;
        isInSubProgramScope = false;
    }
  :
  PROCEDURE des1=designator
  {
    subp = new FVhPSubProgramDecl( $des1.tok.getText(), $des1.tok.getLine() );
    FVhPDescription.instance().pushScope(subp);   
  }
  (
      LPAREN intf1=formal_parameter_list
      { 
          intf_ele = $intf1.intf_ele;
      } 
      RPAREN
  )?
  |
  (
    PURE
    | IMPURE
  )?
  FUNCTION des2=designator
  {
      String fname = $des2.tok.getText();
      subp = new FVhPSubProgramDecl( $des2.tok.getText(), $des2.tok.getLine() );
      subp.setFunction(true);
      FVhPDescription.instance().pushScope(subp);
  } 
  (
      LPAREN intf2=formal_parameter_list
      { 
          intf_ele = $intf2.intf_ele;
      } 
      RPAREN
  )? 
  RETURN name
  {
      String fname = $des2.tok.getText();
      String rettype = $name.tok.getText();
      FVhPBase typemark = FVhPDescription.instance().getSymbol( rettype );
      if ( null == typemark )
    {
        LogMessages.instance().print(3561, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+
                                      $name.tok.getText() + 
                                      "' in " + FVhPDescription.instance().getFile() +
                                      ":" + $name.tok.getLine() );
    }
      FVhPSubTypeIndication ret =  new FVhPSubTypeIndication($name.tok.getLine(), typemark, $name.rangelist);
      if ( null != subp )
      {      
        subp.setReturnType(ret);
      }
  }
  ;

subprogram_statement_part
  :
  (sequential_statement)*
  ;

subtype_declaration returns [FVhPBase subtypedecl]
  :
  SUBTYPE identifier IS subtype_indication SEMI
  {
      $subtypedecl = new FVhPSubTypeDecl($identifier.tok.getText(), 
                                         $subtype_indication.sind, 
                                         $identifier.tok.getLine());
      FVhPDescription.instance().addDeclaration( $subtypedecl );
      FVhPDescription.instance().addSymbol($identifier.tok.getText(), $subtypedecl );
  }
  ;

// VHDLAMS 1076.1-1999 declares first name as optional. Here, second name
// is made optional to prevent antlr nondeterminism.
// (9.2.2004, e.f.)

subtype_indication returns [FVhPSubTypeIndication sind]
    @init
    {
        FVhPBase cons=null;
        String resolutionFuncName = null;
        String objName = null;
    }
  :
  nm1=name 
  (
      nm2=name
      {
          resolutionFuncName = getNameString($nm1.obj);
      }
  )? 
  (
      constraint 
      {
          cons = $constraint.cons;
      }
  )?
  {
      LinkedList<FVhPBase> ranges = $nm1.rangelist;
      if ( null != $nm1.rangelist && null !=  cons )
      {
          LogMessages.instance().print(3560, LogMessages.MSG_SEVERITY.ERROR, "Failed to process array type declaration " +
                                      "with multiple dimensions ...");
      }
      else if ( null != cons )
      {
          ranges = new LinkedList<FVhPBase>();
          ranges.add( cons );
      }
      FVhPBase typemark = null;
      FVhPBase resolveFunc = null;
      String typeName = null;
      if ( null != resolutionFuncName )
      {
        typeName = getNameString( $nm2.obj );
        
        typemark = FVhPDescription.instance().getSymbol( typeName );
        if ( null == typemark && $nm2.obj instanceof FVhPSelectedName )
        {
          typemark = $nm2.obj;
        }
        resolveFunc =  FVhPDescription.instance().getSymbol(resolutionFuncName);
      }
      else
      {
        typeName = getNameString( $nm1.obj );
        if ( null == typeName )
        {
            typeName = $nm1.tok.getText();
        }
        if ( null != typeName )
        {
            typemark = FVhPDescription.instance().getSymbol( typeName );
        }
        if ( null == typemark && $nm1.obj instanceof FVhPSelectedName )
        {
            typemark = $nm1.obj;
        }
      }
      if ( null == typemark )
      {
        LogMessages.instance().print(3559, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+
                                      $nm1.tok.getText() +
                                      "' in " + FVhPDescription.instance().getFile() +
                                      ":" + $nm1.tok.getLine() );
      }
      $sind = new FVhPSubTypeIndication( $nm1.tok.getLine(), typemark, ranges);
      if ( null != resolveFunc )
      {
          $sind.setResolutionFunc( resolveFunc );
      }     
  }
  ;

suffix returns [Token tok, String all]
    :
    identifier 
    { 
        $tok = $identifier.tok ;
    }
    | CHARACTER_LITERAL
    {
        $tok = $CHARACTER_LITERAL;
    }
    | STRING_LITERAL
    {
        $tok = $STRING_LITERAL;
    }
    | ALL 
    { 
        $all = new String("all");
    }
    ;

target returns [FVhPBase target]
  :
  name
  {
  	  FVhPBase sym = $name.obj;
  	  if ( null == sym )
  	  {
      	sym = FVhPDescription.instance().getSymbol($name.tok.getText());
      }
      if ( null == sym )
      {
        LogMessages.instance().print(3558, LogMessages.MSG_SEVERITY.ERROR,"Found reference to undefined symbol '"+
                                      $name.tok.getText() +
                                      "' in " + FVhPDescription.instance().getFile() +
                                      ":" + $name.tok.getLine() );
      }

      if ( null != $name.rangelist && false == $name.rangelist.isEmpty() )
      {
          if ( null != $name.declaredObj && FVhPBase.ObjType.SIMPLENAME != $name.declaredObj.getObjType() )
          {
           sym = $name.declaredObj;
          }
          $target = new FVhPPartSelect($name.tok.getLine(), sym, $name.rangelist);
      }
      else if ( null != $name.bitsel )
      {
          $target = new FVhPTarget($name.tok.getLine(), $name.bitsel);
      }
      else
      {
          $target = new FVhPTarget($name.tok.getLine(), sym);
      }
  }
  | aggregate
  {
      $target = new FVhPTarget($aggregate.aggr.getLine(), $aggregate.aggr );
  }
  ;

term returns [FVhPExprBase expr]
    @init
    {
        FVhPExprBase lhs = null;
        FVhPExprBase rhs = null;
        FVhPExprBinary.OpType op = FVhPExprBinary.OpType.UNKNOWN;
    }
  :
  f1=factor
  {
      lhs = $f1.expr;
  }
       (options {greedy=true;}: 
           multiplying_operator
           {
               op = $multiplying_operator.op;
           } 
           f2=factor
           {
               rhs = $f2.expr;
           }
       )*
  {
      if ( null != rhs )
      {
          $expr = new FVhPExprBinary ( ( null == lhs ) ? 0 : lhs.getLine(), lhs, rhs, op );
      }
      else
      {
          $expr = lhs;
      }
  }
  ;

terminal_aspect
  :
  nm=name (TO name)?
  {
      LogMessages.instance().print(3557, LogMessages.MSG_SEVERITY.WARNING, "Ignoring the terminal aspect found at " +  
                 FVhPDescription.instance().getFile() + 
                 ":" +  $nm.tok.getLine() + " ...");
  }
  ;


timeout_clause returns [FVhPExprBase expr]
  :
  FOR expression
  {
      $expr = $expression.expr;
  }
  ;


type_declaration
    @init
    {
        FVhPTypeDefBase def = null;
    }
  :
  TYPE id=identifier (IS typedef=type_definition { def = $typedef.type;} )? SEMI
  {
      String name =  $id.tok.getText();
      FVhPTypeDecl decl = new FVhPTypeDecl( name, def, $id.tok.getLine() );
      FVhPDescription desc = FVhPDescription.instance();
      FVhPBase sym = (FVhPBase)desc.getSymbol( name );
      if ( null != sym )
      {
        LogMessages.instance().print(3556, LogMessages.MSG_SEVERITY.ERROR, "Found already defined symbol '" + 
                                       name + "'  in " +
                                       desc.getFile() +
                                       ":" + $id.tok.getLine() );
      } 
      desc.insertSymbol( name, decl );
      desc.addDeclaration(decl);
  }
  ;

type_definition returns [FVhPTypeDefBase type ]
  :
  sc=scalar_type_definition
  {
      $type = $sc.type;
  }
  | 
  cmp=composite_type_definition
  {
      $type = $cmp.def;
  }
  | 
  access_type_definition
  {
  	$type =  $access_type_definition.accesstype;
  }
  | 
  file_type_definition
  {
	$type = $file_type_definition.typedef;
  }
  ;


// NOTE that name is used here, so need to check if result of name is
// a seleceted_name (i.e. prefix DOT suffix).
// (30.3.2004, e.f.)
//use_clause
//  : USE name ( COMMA name )* SEMI
//  ;

use_clause returns [LinkedList<String> uses, LinkedList<FVhPSelectedName> useClauses]
  :
  USE  nm=name 
  {
      $uses = new LinkedList<String>();
      $uses.add($nm.fullname);
      $useClauses = new LinkedList<FVhPSelectedName>();
      $useClauses.add( (FVhPSelectedName)$nm.selectedName );
  } 
  (
      COMMA nm2=name
      {
          $uses.add($nm2.fullname);
          $useClauses.add( (FVhPSelectedName)$nm2.selectedName );
      }
  )* SEMI
  {
      uses = $uses;
//      useClauses = $useClauses; 
  }
  ;

variable_assignment_statement
    @init
    {
        String label = null;
    }
  :
  (label_colon { label = $label_colon.tok.getText();})?  
  target
  {
    setCurrentTarget(  $target.target );
  }
  VARASGN 
  expression SEMI
  {
        FVhPVariableAssignment assmt = new FVhPVariableAssignment($target.target.getLine(), 
                                                  $target.target, 
                                                  $expression.expr);
        assmt.setLabel( label );                                          
        FVhPDescription.instance().addStatement(assmt);
        setCurrentTarget(  null );
  }  
  ;

variable_declaration
    @init
    {
        FVhPExprBase initval = null;
    }
    @after 
    {
        myCurrentSind = null;
    }
  :
  (SHARED)? VARIABLE identifier_list COLON subtype_indication { myCurrentSind = $subtype_indication.sind; } (VARASGN expression{initval=$expression.expr;})? SEMI
    {
      LinkedList<Token> ids = $identifier_list.ids;
       if ( null != ids &&  false == ids.isEmpty() )
     {
        Iterator itr = ids.iterator(); 
        while(itr.hasNext()) 
        {
            Token tok = (Token)itr.next(); 
            String varName = tok.getText();
            
            FVhPIdentifier ident = new FVhPIdentifier( tok.getLine(), varName );
            FVhPVariable var = new FVhPVariable ( tok.getLine(), ident, $subtype_indication.sind );
            FVhPDescription desc = FVhPDescription.instance();
            FVhPBase sym = (FVhPBase)desc.getSymbol( varName );
            if ( null != sym && desc.getScope() == sym.getScope() )
            {
                LogMessages.instance().print(3555, LogMessages.MSG_SEVERITY.ERROR,
                                            "Found already defined symbol '" + 
                                             varName + "' reference in " +
                                             desc.getFile() +
                                             ":" + tok.getLine() +
                                             " in file " + sym.getFile() + ":"  + sym.getLine());
            } 
            desc.insertSymbol( varName, var );
            if ( null != initval )
            {
                initval.setSubtypeIndication( $subtype_indication.sind );
            }
            var.setInitialValue(initval);
            desc.addDeclaration(var);
        } 
    } 
  }  
  ;

wait_statement returns [FVhPBase waitstmt]
    @init
    {
        String label = null;
        FVhPWaitStmt stmt = null;
    }
  :
  (label_colon {label = $label_colon.tok.getText();})? 
  WAIT
  {
      stmt = new FVhPWaitStmt( $WAIT.getLine());
      stmt.setLabel( label );
  } 
  (
      sensitivity_clause
      {
        if ( null != stmt )
        {
          stmt.setSensitivityList( $sensitivity_clause.sensitivity_list );
        }
      }
  )? 
  (
      condition_clause
      {
        if ( null != stmt )
        {
          stmt.setUntilCond( $condition_clause.expr );
        }
      }
  )? 
  (
      timeout_clause
      {
        if ( null != stmt )
        {
          stmt.setForDuration( $timeout_clause.expr );
        }
      }
  )?  
  SEMI
  {
    $waitstmt = stmt;
    FVhPDescription.instance().addStatement ( $waitstmt );
  }  
  ;

waveform returns [FVhPWaveForm wvfrm, Integer line]
    @init
    {
         LinkedList<FVhPWaveFrmEle> waveforms = null;
    }
  :
  ele1=waveform_element
  {
    waveforms = new LinkedList<FVhPWaveFrmEle>();
    waveforms.add( $ele1.ele );
    $line = $ele1.ele.getLine();
  } 
  (COMMA 
      ele2=waveform_element
      {
        if ( null != waveforms )
        {
          waveforms.add( $ele2.ele );
        }
      }
  )*
  {
      $wvfrm = new FVhPWaveForm($ele1.ele.getLine(), waveforms, null );
  }
  | UNAFFECTED
  {
      $wvfrm = new FVhPWaveForm( $UNAFFECTED.getLine(), null, null );
      ((FVhPWaveForm)$wvfrm).setUnaffedted( true );
  }
  ;

waveform_element returns [FVhPWaveFrmEle ele]
  :
  e1=expression 
  {
      $ele = new FVhPWaveFrmEle($e1.expr.getLine(), $e1.expr, null);
  }
  (AFTER 
      e2=expression
      {
          $ele.setAfterClause($e2.expr);
      }
  )?
  ;

ABS
  :
  'abs'
  ;

ACCESS
  :
  'access'
  ;

AFTER
  :
  'after'
  ;

ALIAS
  :
  'alias'
  ;

ALL
  :
  'all'
  ;

AND
  :
  'and'
  ;

ARCHITECTURE
  :
  'architecture'
  ;

ARRAY
  :
  'array'
  ;

ASSERT
  :
  'assert'
  ;

ATTRIBUTE
  :
  'attribute'
  ;

BEGIN
  :
  'begin'
  ;

BLOCK
  :
  'block'
  ;

BODY
  :
  'body'
  ;

BUFFER
  :
  'buffer'
  ;

BUS
  :
  'bus'
  ;

CASE
  :
  'case'
  ;

COMPONENT
  :
  'component'
  ;

CONFIGURATION
  :
  'configuration'
  ;

CONSTANT
  :
  'constant'
  ;

DISCONNECT
  :
  'disconnect'
  ;

DOWNTO
  :
  'downto'
  ;

ELSE
  :
  'else'
  ;

ELSIF
  :
  'elsif'
  ;

END
  :
  'end'
  ;

ENTITY
  :
  'entity'
  ;

EXIT
  :
  'exit'
  ;

FILE
  :
  'file'
  ;

FOR
  :
  'for'
  ;

FUNCTION
  :
  'function'
  ;

GENERATE
  :
  'generate'
  ;

GENERIC
  :
  'generic'
  ;

GROUP
  :
  'group'
  ;

GUARDED
  :
  'guarded'
  ;

IF
  :
  'if'
  ;

IMPURE
  :
  'impure'
  ;

IN
  :
  'in'
  ;

INERTIAL
  :
  'inertial'
  ;

INOUT
  :
  'inout'
  ;

IS
  :
  'is'
  ;

LABEL
  :
  'label'
  ;

LIBRARY
  :
  'library'
  ;

LINKAGE
  :
  'linkage'
  ;

LITERAL
  :
  'literal'
  ;

LOOP
  :
  'loop'
  ;

MAP
  :
  'map'
  ;

MOD
  :
  'mod'
  ;

NAND
  :
  'nand'
  ;

NEW
  :
  'new'
  ;

NEXT
  :
  'next'
  ;

NOR
  :
  'nor'
  ;

NOT
  :
  'not'
  ;

NULLTOK
  :
  'null'
  ;
//  NULL : 'null';

OF
  :
  'of'
  ;

ON
  :
  'on'
  ;

OPEN
  :
  'open'
  ;

OR
  :
  'or'
  ;

OTHERS
  :
  'others'
  ;

OUT
  :
  'out'
  ;

PACKAGE
  :
  'package'
  ;

PORT
  :
  'port'
  ;

POSTPONED
  :
  'postponed'
  ;

PROCEDURE
  :
  'procedure'
  ;

PROCESS
  :
  'process'
  ;

PURE
  :
  'pure'
  ;

RANGETOK
  :
  'range'
  ;

RECORD
  :
  'record'
  ;

REGISTER
  :
  'register'
  ;

REJECT
  :
  'reject'
  ;

REM
  :
  'rem'
  ;

REPORT
  :
  'report'
  ;

RETURN
  :
  'return'
  ;

ROL
  :
  'rol'
  ;

ROR
  :
  'ror'
  ;

SELECT
  :
  'select'
  ;

SEVERITY
  :
  'severity'
  ;

SIGNAL
  :
  'signal'
  ;

SHARED
  :
  'shared'
  ;

SLA
  :
  'sla'
  ;

SLL
  :
  'sll'
  ;

SRA
  :
  'sra'
  ;

SRL
  :
  'srl'
  ;

SUBTYPE
  :
  'subtype'
  ;

THEN
  :
  'then'
  ;

TO
  :
  'to'
  ;

TRANSPORT
  :
  'transport'
  ;

TYPE
  :
  'type'
  ;

UNAFFECTED
  :
  'unaffected'
  ;

UNITS
  :
  'units'
  ;

UNTIL
  :
  'until'
  ;

USE
  :
  'use'
  ;

VARIABLE
  :
  'variable'
  ;

WAIT
  :
  'wait'
  ;

WHEN
  :
  'when'
  ;

WHILE
  :
  'while'
  ;

WITH
  :
  'with'
  ;

XNOR
  :
  'xnor'
  ;

XOR
  :
  'xor'
  ;

TAB
  :
  ('\t')+ { $channel = HIDDEN;}
  ;

SPACE
  :
  (' ')+ { $channel = HIDDEN;}
  ;

NEWLINE
  :
  '\n' { $channel = HIDDEN;}
  ;

COMMENT
  :
  '--' (~'\n')* { $channel = HIDDEN;}
  ;

DOUBLESTAR
  :
  '**'
  ;

ASSIGN
  :
  '=='
  ;

LE
  :
  '<='
  ;

GE
  :
  '>='
  ;

ARROW
  :
  '=>'
  ;

NEQ
  :
  '/='
  ;

VARASGN
  :
  ':='
  ;

BOX
  :
  '<>'
  ;

DBLQUOTE
  :
  '\"'
  ;

SEMI
  :
  ';'
  ;

COMMA
  :
  ','
  ;

AMPERSAND
  :
  '&'
  ;

APOSTROPHE
  :
  '\''
  (
    (
      (
        'a'..'z'
        | '0'..'9'
        | '&'
        | '\''
        | '('
        | ')'
        | '+' 
        | ','
        | '-'
        | '.'
        | '/'
        | ':'
        | ';'
        | '<'
        | '='
        | '>'
        | '|'
        | ' '
        | OTHER_SPECIAL_CHARACTER
        | '\\'
        | '#'
        | '['
        | ']'
        | '_'
        | '*'
        | '"'
      )
      '\''
    )
      =>
    (
      'a'..'z'
      | '0'..'9'
      | '&'
      | '\''
      | '('
      | ')'
      | '+'  
      | ','
      | '-'
      | '.'
      | '/'
      | ':'
      | ';'
      | '<'
      | '='
      | '>'
      | '|'
      | ' '
      | OTHER_SPECIAL_CHARACTER
      | '\\'
      | '#'
      | '['
      | ']'
      | '_'
      | '*'
      | '"'
    )
    '\'' { $type = CHARACTER_LITERAL;}
  )?
  ;


LPAREN
  :
  '('
  ;

RPAREN
  :
  ')'
  ;

LBRACKET
  :
  '['
  ;

RBRACKET
  :
  ']'
  ;

COLON
  :
  ':'
  ;

MUL
  :
  '*'
  ;

DIV
  :
  '/'
  ;

PLUS
  :
  '+' 
  ;

MINUS
  :
  '-'
  ;

LOWERTHAN
  :
  '<'
  ;

GREATERTHAN
  :
  '>'
  ;

EQ
  :
  '='
  ;

BAR
  :
  '|'
  ;

DOT
  :
  '.'
  ;

BACKSLASH
  :
  '\\'
  ;

BASIC_IDENTIFIER
  :
  'a'..'z'
  (
    'a'..'z'
    | '0'..'9'
    | '_'
  )*
  ;

EXTENDED_IDENTIFIER
  :
  '\\'
  (
    'a'..'z'
    | '0'..'9'
    | '&'
    | '\''
    | '('
    | ')'
    | '+' 
    | ','
    | '-'
    | '.'
    | '/'
    | ':'
    | ';'
    | '<'
    | '='
    | '>'
    | '|'
    | ' '
    | OTHER_SPECIAL_CHARACTER
    | '\\'
    | '#'
    | '['
    | ']'
    | '_'
  )+
  '\\'
  ;

BIT_STRING_LITERAL
  :
  (
    'b'
    | 'o'
    | 'x'
  )
  '\"' BASED_INTEGER '\"'
  ;

//STRING_LITERAL
//  : '\"' ( 'a'..'z' | '0'..'9' | '&' | '\'' | '(' | ')'
//    | '+' | ',' | '-' | '.' | '/' | ':' | ';' | '<' | '=' | '>' | '|'
//    | ' ' | OTHER_SPECIAL_CHARACTER | '\\' | "\"\"" | '*'
//    | '#' | '[' | ']' | '_' )* '\"'
//  ;

STRING_LITERAL
  :
  '"' ( ~('\"' ))* '"'
  ;

// NOTE that BASED_LITERAL is implicitely defined by referencing in the
// parser grammar.
// (2.2.2004, e.f.)

DECIMAL_LITERAL 
  :
  int1=INTEGER 
  (
    (
      ('.' int2=INTEGER)? 
      (EXPONENT)?
    )
    |
    (
      '#' b1=BASED_INTEGER 
      ('.' b2=BASED_INTEGER)? 
      '#' (EXPONENT)?
    )
    { $type = BASED_LITERAL;}
  )
  ;

CR
  :
  '\r' { $channel = HIDDEN;}
  ;

fragment
OTHER_SPECIAL_CHARACTER
  :
  '!'
  | '$'
  | '%'
  | '@'
  | '?'
  | '^'
  | '`'
  | '{'
  | '}'
  | '~'
  | ''
  ;

fragment
INTEGER
  :
  '0'..'9'
  (
    '_'
    | '0'..'9'
  )*
  ;

fragment
EXPONENT
  :
  'e'
  (
    '+' 
    | '-' 
  )?
  INTEGER
  ;

fragment
BASED_INTEGER
  :
  (
    '0'..'9'
    | 'a'..'f'
  )
  (
    '_'
    | '0'..'9'
    | 'a'..'f'
  )*
  ;
